# Copyright 2025 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyGlove typing to JSON schema conversion."""

import dataclasses
import inspect
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from pyglove.core import utils
from pyglove.core.typing import callable_signature
from pyglove.core.typing import class_schema
from pyglove.core.typing import key_specs as ks
from pyglove.core.typing import value_specs as vs


@dataclasses.dataclass
class _DefEntry:
  definition: Optional[Dict[str, Any]]
  num_references: int = 0


def _json_schema_from_schema(
    schema: class_schema.Schema,
    *,
    defs: Dict[str, _DefEntry],
    type_name: Optional[str] = None,
    include_type_name: bool = True,
    include_subclasses: bool = False,
) -> Dict[str, Any]:
  """Converts a PyGlove schema to JSON schema."""
  title = schema.name.split('.')[-1] if schema.name else None
  entry = defs.get(title)
  if entry is None:
    # NOTE(daiyip): Add a forward reference to the entry so recursive
    # references can be resolved.
    entry = _DefEntry(definition=None, num_references=0)
    if title:
      defs[title] = entry

    additional_properties = False
    properties = {}
    required = []

    if type_name and include_type_name:
      properties[utils.JSONConvertible.TYPE_NAME_KEY] = {'const': type_name}
      required.append(utils.JSONConvertible.TYPE_NAME_KEY)

    for key, field in schema.items():
      if isinstance(key, ks.ConstStrKey):
        prop = _json_schema_from_value_spec(
            field.value, defs=defs
        )
        if field.description:
          prop['description'] = field.description
        properties[key.text] = prop
        if not field.value.has_default:
          required.append(key.text)
      else:
        if isinstance(field.value, vs.Any):
          additional_properties = True
        else:
          additional_properties = _json_schema_from_value_spec(
              field.value, defs=defs,
              include_type_name=include_type_name,
              include_subclasses=include_subclasses,
          )
    entry.definition = _dict_with_optional_fields([
        ('type', 'object', None),
        ('title', title, None),
        ('description', schema.description, None),
        ('properties', properties, {}),
        ('required', required, []),
        ('additionalProperties', additional_properties, None),
    ])

  entry.num_references += 1
  if title:
    return {'$ref': f'#/$defs/{title}'}
  else:
    assert entry.definition is not None
    return entry.definition   # pytype: disable=bad-return-type


def _json_schema_from_value_spec(
    value_spec: vs.ValueSpec,
    *,
    defs: Dict[str, _DefEntry],
    include_type_name: bool = True,
    include_subclasses: bool = False,
    ignore_nonable: bool = False,
) -> Dict[str, Any]:
  """Converts a value spec to JSON schema."""
  def _child_json_schema(v: vs.ValueSpec, ignore_nonable: bool = False):
    return _json_schema_from_value_spec(
        v, defs=defs,
        include_type_name=include_type_name,
        include_subclasses=include_subclasses,
        ignore_nonable=ignore_nonable
    )

  if isinstance(value_spec, vs.Bool):
    definition = {
        'type': 'boolean'
    }
  elif isinstance(value_spec, vs.Int):
    definition = _dict_with_optional_fields([
        ('type', 'integer', None),
        ('minimum', value_spec.min_value, None),
        ('maximum', value_spec.max_value, None),
    ])
  elif isinstance(value_spec, vs.Float):
    definition = _dict_with_optional_fields([
        ('type', 'number', None),
        ('minimum', value_spec.min_value, None),
        ('maximum', value_spec.max_value, None),
    ])
  elif isinstance(value_spec, vs.Str):
    definition = _dict_with_optional_fields([
        ('type', 'string', None),
        ('pattern', getattr(value_spec.regex, 'pattern', None), None)
    ])
  elif isinstance(value_spec, vs.Enum):
    for v in value_spec.values:
      if not isinstance(v, (str, bool, int, float, type(None))):
        raise ValueError(
            f'Enum candidate {v!r} is not supported for JSON schema generation.'
        )
    definition = {
        'enum': [v for v in value_spec.values if v is not None],
    }
  elif isinstance(value_spec, vs.List):
    definition = {
        'type': 'array',
    }
    if not isinstance(value_spec.element.value, vs.Any):
      definition['items'] = _child_json_schema(value_spec.element.value)
  elif isinstance(value_spec, vs.Dict):
    if value_spec.schema is None:
      definition = {
          'type': 'object',
          'additionalProperties': True
      }
    else:
      definition = _json_schema_from_schema(
          value_spec.schema, defs=defs, include_type_name=include_type_name
      )
  elif isinstance(value_spec, vs.Object):
    def _json_schema_from_cls(cls: type[Any]):
      schema = getattr(cls, '__schema__', None)
      if schema is None:
        schema = callable_signature.signature(cls).to_schema()
      return _json_schema_from_schema(
          schema,
          defs=defs,
          type_name=getattr(cls, '__type_name__', None),
          include_type_name=include_type_name,
          include_subclasses=include_subclasses,
      )
    definitions = [_json_schema_from_cls(value_spec.cls)]

    if include_subclasses:
      for subclass in value_spec.cls.__subclasses__():
        if not inspect.isabstract(subclass):
          definitions.append(_json_schema_from_cls(subclass))

    if len(definitions) == 1:
      definition = definitions[0]
    else:
      definition = {'anyOf': _normalize_anyofs(definitions)}
  elif isinstance(value_spec, vs.Union):
    definition = {
        'anyOf': _normalize_anyofs([
            _child_json_schema(v, ignore_nonable=True)
            for v in value_spec.candidates
        ])
    }
  elif isinstance(value_spec, vs.Any):
    # Consider using return {}
    definition = {
        'anyOf': [
            _child_json_schema(v) for v in [
                vs.Bool(),
                vs.Float(),
                vs.Str(),
                vs.List(vs.Any()),
                vs.Dict(),
            ]
        ]
    }
  else:
    raise TypeError(
        f'Value spec {value_spec!r} cannot be converted to JSON schema.'
    )

  if (value_spec.has_default
      and value_spec.default is not None
      and not isinstance(value_spec, vs.Dict)):
    default = utils.to_json(value_spec.default)
    if not include_type_name:
      def _remove_type_name(_, v: Dict[str, Any]):
        if isinstance(v, dict):
          v.pop(utils.JSONConvertible.TYPE_NAME_KEY, None)
        return v
      default = utils.transform(default, _remove_type_name)
    definition['default'] = default

  if not ignore_nonable and value_spec.is_noneable:
    nullable = {'type': 'null'}
    if 'anyOf' in definition:
      definition['anyOf'].append(nullable)
    else:
      definition = {'anyOf': [definition, nullable]}
    if value_spec.default is None:
      definition['default'] = None
  return definition


def _normalize_anyofs(candidates: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
  flattened = []
  for candidate in candidates:
    if 'anyOf' in candidate:
      flattened.extend(_normalize_anyofs(candidate['anyOf']))
    else:
      flattened.append(candidate)
  return flattened


def _dict_with_optional_fields(
    pairs: List[Tuple[str, Any, Any]]
) -> Dict[str, Any]:
  return {
      k: v for k, v, default in pairs if v != default
  }


def to_json_schema(
    value: Union[vs.ValueSpec, class_schema.Schema, Any],
    *,
    include_type_name: bool = True,
    include_subclasses: bool = False,
    inline_nested_refs: bool = False,
) -> Dict[str, Any]:
  """Converts a value spec to JSON schema."""
  defs = dict()
  if isinstance(value, class_schema.Schema):
    root = _json_schema_from_schema(
        value, defs=defs,
        type_name=value.name,
        include_type_name=include_type_name,
        include_subclasses=include_subclasses,
    )
  else:
    value = vs.ValueSpec.from_annotation(value, auto_typing=True)
    root = _json_schema_from_value_spec(
        value,
        defs=defs,
        include_type_name=include_type_name,
        include_subclasses=include_subclasses,
    )
  return _canonicalize_schema(root, defs, inline_nested_refs=inline_nested_refs)


def _canonicalize_schema(
    root: Dict[str, Any],
    defs: Dict[str, _DefEntry],
    *,
    inline_nested_refs: bool = True,
    include_defs: bool = True,
) -> Dict[str, Any]:
  """Canonicalizes a JSON schema."""
  if not defs:
    return root

  def _maybe_inline_ref(k: utils.KeyPath, v: Any):
    del k
    if isinstance(v, dict) and '$ref' in v:
      ref_key = v['$ref'].split('/')[-1]
      if defs[ref_key].num_references == 1:
        defs[ref_key].num_references -= 1
        if not inline_nested_refs:
          return defs[ref_key].definition
        else:
          return _canonicalize_schema(
              defs[ref_key].definition, defs=defs, include_defs=False,
              inline_nested_refs=True
          )
    return v

  new_root = utils.transform(root, _maybe_inline_ref)
  if not include_defs:
    return new_root
  referenced_defs = {
      k: v.definition for k, v in defs.items() if v.num_references > 0
  }
  canonical_form = {'$defs': referenced_defs} if referenced_defs else {}
  canonical_form.update(new_root)
  if 'default' in root:
    canonical_form['default'] = root['default']
  return canonical_form

#
# Override to_json_schema() method for ValueSpec and Schema.
#


def _value_spec_to_json_schema(
    self: Union[vs.ValueSpec, class_schema.Schema, Any],
    *,
    include_type_name: bool = True,
    include_subclasses: bool = False,
    inline_nested_refs: bool = False,
    **kwargs
) -> Dict[str, Any]:
  """Converts a value spec to JSON schema."""
  return to_json_schema(
      self,
      include_type_name=include_type_name,
      include_subclasses=include_subclasses,
      inline_nested_refs=inline_nested_refs,
      **kwargs
  )


def _schema_to_json_schema(
    self: class_schema.Schema,
    *,
    include_type_name: bool = True,
    include_subclasses: bool = False,
    inline_nested_refs: bool = False,
    **kwargs
) -> Dict[str, Any]:
  """Converts a schema to JSON schema."""
  return to_json_schema(
      self,
      include_type_name=include_type_name,
      include_subclasses=include_subclasses,
      inline_nested_refs=inline_nested_refs,
      **kwargs
  )

class_schema.ValueSpec.to_json_schema = _value_spec_to_json_schema
class_schema.Schema.to_json_schema = _schema_to_json_schema

#
# from JSON schema to PyGlove value spec.
#


def _json_schema_to_value_spec(
    json_schema: Dict[str, Any],
    defs: Dict[str, Type[Any]],
    class_fn: Optional[Callable[[str, class_schema.Schema], Type[Any]]],
    add_json_schema_as_metadata: bool,
) -> class_schema.ValueSpec:
  """Converts a JSON schema to a value spec."""
  # Generate code to convert JSON schema to value spec.
  def _value_spec(value_schema: Dict[str, Any]):
    return _json_schema_to_value_spec(
        value_schema, defs, class_fn, add_json_schema_as_metadata
    )

  if '$ref' in json_schema:
    # TODO(daiyip): Support circular references.
    ref_key = json_schema['$ref'].split('/')[-1]
    type_ref = defs.get(ref_key)
    if type_ref is None:
      raise ValueError(
          f'Reference {ref_key!r} not defined in defs. '
          'Please make sure classes being referenced are defined '
          'before the referencing classes. '
      )
    return type_ref
  type_str = json_schema.get('type')
  default = json_schema.get('default', utils.MISSING_VALUE)
  if type_str is None:
    if 'enum' in json_schema:
      for v in json_schema['enum']:
        if not isinstance(v, (str, int, float, bool)):
          raise ValueError(
              f'Enum candidate {v!r} is not supported for JSON schema '
              'conversion.'
          )
      return vs.Enum(
          default,
          [v for v in json_schema['enum'] if v is not None]
      )
    elif 'anyOf' in json_schema:
      candidates = []
      accepts_none = False
      for v in json_schema['anyOf']:
        candidate = _value_spec(v)
        if candidate.frozen and candidate.default is None:
          accepts_none = True
          continue
        candidates.append(candidate)

      if len(candidates) == 1:
        spec = candidates[0]
      else:
        spec = vs.Union(candidates)
      if accepts_none:
        spec = spec.noneable()
      return spec
  elif type_str == 'null':
    return vs.Any().freeze(None)
  elif type_str == 'boolean':
    return vs.Bool(default=default)
  elif type_str == 'integer':
    minimum = json_schema.get('minimum')
    maximum = json_schema.get('maximum')
    return vs.Int(min_value=minimum, max_value=maximum, default=default)
  elif type_str == 'number':
    minimum = json_schema.get('minimum')
    maximum = json_schema.get('maximum')
    return vs.Float(min_value=minimum, max_value=maximum, default=default)
  elif type_str == 'string':
    pattern = json_schema.get('pattern')
    return vs.Str(regex=pattern, default=default)
  elif type_str == 'array':
    items = json_schema.get('items')
    return vs.List(_value_spec(items) if items else vs.Any(), default=default)
  elif type_str == 'object':
    schema = _json_schema_to_schema(
        json_schema, defs, class_fn, add_json_schema_as_metadata
    )
    if class_fn is not None and 'title' in json_schema:
      return vs.Object(class_fn(json_schema['title'], schema))
    return vs.Dict(schema=schema if schema.fields else None)
  raise ValueError(f'Unsupported type {type_str!r} in JSON schema.')


def _json_schema_to_schema(
    json_schema: Dict[str, Any],
    defs: Dict[str, Type[Any]],
    class_fn: Optional[Callable[[str, class_schema.Schema], Type[Any]]],
    add_json_schema_as_metadata: bool,
) -> class_schema.Schema:
  """Converts a JSON schema to a schema."""
  title = json_schema.get('title')
  properties = json_schema.get('properties', {})
  fields = []
  required = set(json_schema.get('required', []))
  for name, property_schema in properties.items():
    value_spec = _json_schema_to_value_spec(
        property_schema, defs, class_fn, add_json_schema_as_metadata
    )
    if name not in required and not value_spec.has_default:
      value_spec = value_spec.noneable()
    fields.append(
        class_schema.Field(
            name,
            value_spec,
            description=property_schema.get('description'),
            metadata=(
                dict(json_schema=property_schema)
                if add_json_schema_as_metadata else None
            )
        )
    )
  additional_properties = json_schema.get('additionalProperties')
  if additional_properties:
    if isinstance(additional_properties, dict):
      value_spec = _json_schema_to_value_spec(
          additional_properties, defs, class_fn, add_json_schema_as_metadata
      )
    else:
      value_spec = vs.Any()
    fields.append(class_schema.Field(ks.StrKey(), value_spec))
  return class_schema.Schema(
      name=title,
      description=json_schema.get('description'),
      fields=fields,
      allow_nonconst_keys=True,
  )


@classmethod
def _value_spec_from_json_schema(
    cls,
    json_schema: Dict[str, Any],
    class_fn: Optional[Callable[[str, class_schema.Schema], Type[Any]]] = None,
    add_json_schema_as_metadata: bool = False,
) -> class_schema.ValueSpec:
  """Creates a PyGlove value spec from a JSON schema.

  Args:
    json_schema: The JSON schema for a value spec.
    class_fn: A function that creates a PyGlove class from a class name and a
      schema. If None, all "object" type properties will be converted to
      `pg.typing.Dict`. Otherwise, "object" type properties will be converted to
      a class.
    add_json_schema_as_metadata: Whether to add the JSON schema as field
      metadata.

  Returns:
    A PyGlove value spec.
  """
  del cls
  defs = {}
  if '$defs' in json_schema:
    for key, def_entry in json_schema['$defs'].items():
      defs[key] = _json_schema_to_value_spec(
          def_entry, defs, class_fn, add_json_schema_as_metadata
      )
  return _json_schema_to_value_spec(
      json_schema, defs, class_fn, add_json_schema_as_metadata
  )


@classmethod
def _schema_from_json_schema(
    cls,
    json_schema: Dict[str, Any],
    class_fn: Optional[Callable[[str, class_schema.Schema], Type[Any]]] = None,
    add_json_schema_as_metadata: bool = False,
) -> class_schema.Schema:
  """Creates a PyGlove schema from a JSON schema.

  Args:
    json_schema: The JSON schema to convert.
    class_fn: A function that creates a PyGlove class from a class name and a
      schema. If None, all "object" type properties will be converted to
      `pg.typing.Dict`. Otherwise, "object" type properties will be converted to
      a class.
    add_json_schema_as_metadata: Whether to add the JSON schema as field
      metadata.

  Returns:
    A PyGlove schema.
  """
  del cls
  if json_schema.get('type') != 'object':
    raise ValueError(
        f'JSON schema is not an object type: {json_schema!r}'
    )
  return _json_schema_to_schema(
      json_schema, {}, class_fn, add_json_schema_as_metadata
  )


class_schema.ValueSpec.from_json_schema = _value_spec_from_json_schema
class_schema.Schema.from_json_schema = _schema_from_json_schema



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface and functions for JSON conversion."""

import abc
import base64
import collections
import contextlib
import importlib
import inspect
import marshal
import pickle
import types
import typing
from typing import Any, Callable, ContextManager, Dict, Iterable, Iterator, List, Optional, Sequence, Set, Tuple, Type, TypeVar, Union

# Nestable[T] is a (maybe) nested structure of T, which could be T, a Dict
# a List or a Tuple of Nestable[T]. We use a Union to fool PyType checker to
# make Nestable[T] a valid type annotation without type check.
T = TypeVar('T')
Nestable = Union[Any, T]  # pytype: disable=not-supported-yet

# pylint: disable=invalid-name
JSONPrimitiveType = Union[int, float, bool, str]

# pytype doesn't support recursion. Use Any instead of 'JSONValueType'
# in List and Dict.
JSONListType = List[Any]
JSONDictType = Dict[Union[str, int], Any]
JSONValueType = Union[JSONPrimitiveType, JSONListType, JSONDictType]

# pylint: enable=invalid-name


class _TypeRegistry:
  """A registry for mapping a string name to type definition.

  This class is used for looking up type definition by a string identifier for
  deserialization.
  """

  def __init__(self):
    """Constructor."""
    # NOTE(daiyip): the order of keys in the dict is preserved. As a result,
    # in `pg.wrapping.apply_wrappers`, the latest registered wrapper
    # class will always be picked up when there are multiple wrapper classes
    # registered for a user class.
    self._type_to_cls_map = dict()
    self._prefix_mapping = dict()
    self._ondemand_registry_stack = []

  def register(
      self, type_name: str, cls: Type[Any], override_existing: bool = False
      ) -> None:
    """Register a ``symbolic.Object`` class with a type name.

    Args:
      type_name: String identifier for the class, which will be used as the
        value of `_type` property when deciding which class to construct object
        when converting a JSON value to object.
      cls: Class to register.
      override_existing: Whether allow to override existing value if type name
        is already registered.

    Raises:
      KeyError: If type_name is already registered and override_existing is set
        to False.
    """
    if type_name in self._type_to_cls_map and not override_existing:
      raise KeyError(
          f'Type {type_name!r} has already been registered with class '
          f'{self._type_to_cls_map[type_name].__name__}.')
    self._type_to_cls_map[type_name] = cls

  def add_module_alias(
      self,
      module: str,
      alias: Union[str, Sequence[str]]
  ) -> None:
    """Maps a module name to another name. Usually due to rename."""
    if isinstance(alias, str):
      alias = [alias]
    for name in alias:
      self._prefix_mapping[name] = module

  def is_registered(self, type_name: str) -> bool:
    """Returns whether a type name is registered."""
    return type_name in self._type_to_cls_map

  @contextlib.contextmanager
  def load_types_for_deserialization(
      self,
      *types_to_deserialize: Type[Any],
  ) -> Iterator[Dict[str, Type[Any]]]:
    """Context manager for loading unregistered types for deserialization."""
    if self._ondemand_registry_stack:
      stack_top = dict(self._ondemand_registry_stack[-1])
    else:
      stack_top = {}
    stack_top.update({t.__name__: t for t in types_to_deserialize})
    try:
      self._ondemand_registry_stack.append(stack_top)
      yield stack_top
    finally:
      self._ondemand_registry_stack.pop()

  def class_from_typename(
      self, type_name: str) -> Optional[Type[Any]]:
    """Get class from type name."""
    if self._ondemand_registry_stack:
      top_registry = self._ondemand_registry_stack[-1]
      class_name = type_name.split('.')[-1]
      if class_name in top_registry:
        return top_registry[class_name]

    cls = self._type_to_cls_map.get(type_name, None)
    if cls is None:
      # Modules could be renamed, to load legacy serialized objects, we
      # use prefix mapping to get to their latest registry.
      for k, v in self._prefix_mapping.items():
        if type_name.startswith(f'{k}.'):
          remapped_type_name = type_name.replace(k, v)
          cls = self._type_to_cls_map.get(remapped_type_name, None)
          if cls is not None:
            break
    return cls

  def iteritems(self) -> Iterable[Tuple[str, Type[Any]]]:
    """Iterate type registry."""
    return self._type_to_cls_map.items()


class JSONConvertible(metaclass=abc.ABCMeta):
  """Interface for classes whose instances are convertible from/to JSON.

  A JSON convertible object is an object that can be converted into plain Python
  objects, hence can be serialized into or deserialized from JSON.

  Subclasses of ``JSONConvertible`` should implement:

    * ``to_json``: A method that returns a plain Python dict with a `_type`
      property whose value should identify the class.
    * ``from_json``: A class method that takes a plain Python dict and returns
      an instance of the class.

  Example::

    class MyObject(pg.JSONConvertible):

      def __init__(self, x: int):
        self.x = x

      def to_json(self, **kwargs):
        return {
          '_type': 'MyObject',
          'x': self.x
        }

      @classmethod
      def from_json(cls, json_value, **kwargs):
        return cls(json_value['x'])

  All symbolic types (see :class:`pyglove.Symbolic`) are JSON convertible.
  """

  # Registry for looking up the type definition for a string identifier during
  # deserialization. One key can be used for only one type, while the same type
  # can be registered with many different string identifiers, which can be
  # useful to allow backward compatibility of existing serialized strings.
  _TYPE_REGISTRY = _TypeRegistry()

  # Key in serialized dict that represents the class to restore.
  TYPE_NAME_KEY = '_type'

  # Marker (as the first element of a list) for serializing tuples.
  TUPLE_MARKER = '__tuple__'

  # Type converter that converts a complex type to basic JSON value type.
  # When this field is set by users, the converter will be invoked when a
  # complex value cannot be serialized by existing methods.
  TYPE_CONVERTER: Optional[
      Callable[[Type[Any]], Callable[[Any], JSONValueType]]] = None

  # Class property that indicates whether to automatically register class
  # for deserialization. Subclass can override.
  auto_register = True

  @classmethod
  def from_json(cls, json_value: JSONValueType, **kwargs) -> 'JSONConvertible':
    """Creates an instance of this class from a plain Python value.

    NOTE(daiyip): ``pg.Symbolic`` overrides ``from_json`` class method.

    Args:
      json_value: JSON value type.
      **kwargs: Keyword arguments as flags to control object creation.

    Returns:
      An instance of cls.
    """
    assert isinstance(json_value, dict)
    init_args = {k: from_json(v, **kwargs) for k, v in json_value.items()
                 if k != JSONConvertible.TYPE_NAME_KEY}
    return cls(**init_args)

  @abc.abstractmethod
  def to_json(self, **kwargs) -> JSONValueType:
    """Returns a plain Python value as a representation for this object.

    A plain Python value are basic python types that can be serialized into
    JSON, e.g: ``bool``, ``int``, ``float``, ``str``, ``dict`` (with string
    keys), ``list``, ``tuple`` where the container types should have plain
    Python values as their values.

    Args:
      **kwargs: Keyword arguments as flags to control JSON conversion.

    Returns:
      A plain Python value.
    """

  @classmethod
  def register(
      cls,
      type_name: str,
      subclass: Type['JSONConvertible'],
      override_existing: bool = False
      ) -> None:
    """Registers a class with a type name.

    The type name will be used as the key for class lookup during
    deserialization. A class can be registered with multiple type names, but
    a type name should be uesd only for one class.

    Args:
      type_name: A global unique string identifier for subclass.
      subclass: A subclass of JSONConvertible.
      override_existing: If True, override the class if the type name is
        already present in the registry. Otherwise an error will be raised.
    """
    cls._TYPE_REGISTRY.register(type_name, subclass, override_existing)

  @classmethod
  def add_module_alias(
      cls,
      module: str,
      alias: Union[str, Sequence[str]]
  ) -> None:
    """Adds a module alias so previous serialized objects could be loaded."""
    cls._TYPE_REGISTRY.add_module_alias(module, alias)

  @classmethod
  def is_registered(cls, type_name: str) -> bool:
    """Returns True if a type name is registered. Otherwise False."""
    return cls._TYPE_REGISTRY.is_registered(type_name)

  @classmethod
  def class_from_typename(
      cls, type_name: str) -> Optional[Type['JSONConvertible']]:
    """Gets the class for a registered type name.

    Args:
      type_name: A string as the global unique type identifier for requested
        class.

    Returns:
      A type object if registered, otherwise None.
    """
    return cls._TYPE_REGISTRY.class_from_typename(type_name)

  @classmethod
  def registered_types(cls) -> Iterable[Tuple[str, Type['JSONConvertible']]]:
    """Returns an iterator of registered (serialization key, class) tuples."""
    return cls._TYPE_REGISTRY.iteritems()

  @classmethod
  def load_types_for_deserialization(
      cls,
      *types_to_deserialize: Type[Any]
      ) -> ContextManager[Dict[str, Type[Any]]]:
    """Context manager for loading unregistered types for deserialization.

    Example::

      class A(pg.Object):
        auto_register = False
        x: int

      class B(A):
        y: str
      with pg.JSONConvertile.load_types_for_deserialization(A, B):
          pg.from_json_str(A(1).to_json_str())
          pg.from_json_str(B(1, 'hi').to_json_str())

    Args:
      *types_to_deserialize: A list of types to be loaded for deserialization.

    Returns:
      A context manager within which the objects of the requested types
        could be deserialized.
    """
    return cls._TYPE_REGISTRY.load_types_for_deserialization(
        *types_to_deserialize
    )

  @classmethod
  def to_json_dict(
      cls,
      fields: Dict[str, Union[Tuple[Any, Any], Any]],
      *,
      exclude_default=False,
      exclude_keys: Optional[Set[str]] = None,
      **kwargs) -> Dict[str, JSONValueType]:
    """Helper method to create JSON dict from class and field."""
    json_dict = {JSONConvertible.TYPE_NAME_KEY: _serialization_key(cls)}
    exclude_keys = exclude_keys or set()
    if exclude_default:
      for k, (v, default) in fields.items():
        if k not in exclude_keys and v != default:
          json_dict[k] = to_json(v, **kwargs)
    else:
      json_dict.update(
          {k: to_json(v, **kwargs) for k, v in fields.items()
           if k not in exclude_keys})
    return json_dict

  def __init_subclass__(cls):
    super().__init_subclass__()
    if not inspect.isabstract(cls) and cls.auto_register:
      type_name = _serialization_key(cls)
      JSONConvertible.register(type_name, cls, override_existing=True)


def _serialization_key(
    type_or_function: Union[Type[Any], types.FunctionType]) -> str:
  """Returns the ID for a type or function."""
  serializaton_key = getattr(type_or_function, '__serialization_key__', None)
  if serializaton_key is not None:
    return serializaton_key
  return _type_name(type_or_function)


def _type_name(
    type_or_function: Union[Type[Any], types.FunctionType]) -> str:
  return f'{type_or_function.__module__}.{type_or_function.__qualname__}'


class _OpaqueObject(JSONConvertible):
  """An JSON converter for opaque Python objects."""

  def __init__(self, value: Any, encoded: bool = False):
    if encoded:
      value = self.decode(value)
    self._value = value

  @property
  def value(self) -> Any:
    """Returns the decoded value."""
    return self._value

  def encode(self, value: Any) -> JSONValueType:
    try:
      return base64.encodebytes(pickle.dumps(value)).decode('utf-8')
    except Exception as e:
      raise ValueError(
          f'Cannot encode opaque object {value!r} with pickle.') from e

  def decode(self, json_value: JSONValueType) -> Any:
    assert isinstance(json_value, str), json_value
    try:
      return pickle.loads(base64.decodebytes(json_value.encode('utf-8')))
    except Exception as e:
      raise ValueError('Cannot decode opaque object with pickle.') from e

  def to_json(self, **kwargs) -> JSONValueType:
    return self.to_json_dict({
        'value': self.encode(self._value)
    }, **kwargs)

  @classmethod
  def from_json(cls, json_value: JSONValueType, *args, **kwargs) -> Any:
    del args, kwargs
    assert isinstance(json_value, dict) and 'value' in json_value, json_value
    encoder = cls(json_value['value'], encoded=True)
    return encoder.value


def registered_types() -> Iterable[Tuple[str, Type[JSONConvertible]]]:
  """Returns an iterator of registered (serialization key, class) tuples."""
  return JSONConvertible.registered_types()


def to_json(value: Any, **kwargs) -> Any:
  """Serializes a (maybe) JSONConvertible value into a plain Python object.

  Args:
    value: value to serialize. Applicable value types are:

      * Builtin python types: None, bool, int, float, string;
      * JSONConvertible types;
      * List types;
      * Tuple types;
      * Dict types.

    **kwargs: Keyword arguments to pass to value.to_json if value is
      JSONConvertible.

  Returns:
    JSON value.
  """
  if isinstance(value, (type(None), bool, int, float, str)):
    v = value
  elif isinstance(value, JSONConvertible):
    v = value.to_json(**kwargs)
  elif isinstance(value, tuple):
    v = [JSONConvertible.TUPLE_MARKER] + to_json(list(value), **kwargs)
  elif isinstance(value, list):
    v = [to_json(item, **kwargs) for item in value]
  elif isinstance(value, dict):
    v = {k: to_json(v, **kwargs) for k, v in value.items()}
  elif isinstance(value, (type, typing.GenericAlias)):  # pytype: disable=module-attr
    v = _type_to_json(value)
  elif inspect.isbuiltin(value):
    v = _builtin_function_to_json(value)
  elif inspect.isfunction(value):
    v = _function_to_json(value)
  elif inspect.ismethod(value):
    v = _method_to_json(value)
  # pytype: disable=module-attr
  elif isinstance(value, typing._Final):  # pylint: disable=protected-access
    # pytype: enable=module-attr
    v = _annotation_to_json(value)
  elif value is ...:
    v = {JSONConvertible.TYPE_NAME_KEY: 'type', 'name': 'builtins.Ellipsis'}
  else:
    v, converted = None, False
    if JSONConvertible.TYPE_CONVERTER is not None:
      converter = JSONConvertible.TYPE_CONVERTER(type(value))   # pylint: disable=not-callable
      if converter:
        v = to_json(converter(value))
        converted = True
    if not converted:
      v = _OpaqueObject(value).to_json(**kwargs)
  return v


def from_json(
    json_value: JSONValueType,
    *,
    auto_import: bool = True,
    auto_dict: bool = False,
    **kwargs) -> Any:
  """Deserializes a (maybe) JSONConvertible value from JSON value.

  Args:
    json_value: Input JSON value.
    auto_import: If True, when a '_type' is not registered, PyGlove will
      identify its parent module and automatically import it. For example,
      if the type is 'foo.bar.A', PyGlove will try to import 'foo.bar' and
      find the class 'A' within the imported module.
    auto_dict: If True, dict with '_type' that cannot be loaded will remain
      as dict, with '_type' renamed to 'type_name'.
    **kwargs: Keyword arguments that will be passed to JSONConvertible.__init__.

  Returns:
    Deserialized value.
  """
  typename_resolved = kwargs.pop('_typename_resolved', False)
  if not typename_resolved:
    json_value = resolve_typenames(json_value, auto_import, auto_dict)

  def child_from(v):
    return from_json(v, _typename_resolved=True, **kwargs)

  if isinstance(json_value, list):
    if json_value and json_value[0] == JSONConvertible.TUPLE_MARKER:
      if len(json_value) < 2:
        raise ValueError(
            f'Tuple should have at least one element '
            f'besides \'{JSONConvertible.TUPLE_MARKER}\'. '
            f'Encountered: {json_value}.')
      return tuple([child_from(v) for v in json_value[1:]])
    return [child_from(v) for v in json_value]
  elif isinstance(json_value, dict):
    if JSONConvertible.TYPE_NAME_KEY not in json_value:
      return {k: child_from(v) for k, v in json_value.items()}
    factory_fn = json_value.pop(JSONConvertible.TYPE_NAME_KEY)
    assert factory_fn is not None
    return factory_fn(json_value, **kwargs)
  return json_value


def resolve_typenames(
    json_value: JSONValueType,
    auto_import: bool = True,
    auto_dict: bool = False
) -> JSONValueType:
  """Inplace resolves the "_type" keys with their factories in a JSON tree."""

  def _resolve_typename(v: Dict[str, Any]) -> bool:
    """Returns True if the subtree is resolved for the first time."""
    if JSONConvertible.TYPE_NAME_KEY not in v:
      return True
    if not isinstance(v[JSONConvertible.TYPE_NAME_KEY], str):
      return False
    type_name = v[JSONConvertible.TYPE_NAME_KEY]
    if type_name == 'type':
      factory_fn = _type_from_json
    elif type_name == 'function':
      factory_fn = _function_from_json
    elif type_name == 'method':
      factory_fn = _method_from_json
    else:
      cls = JSONConvertible.class_from_typename(type_name)
      if cls is None:
        if auto_import:
          try:
            cls = _load_symbol(type_name)
            assert inspect.isclass(cls), cls
          except (ModuleNotFoundError, AttributeError) as e:
            if not auto_dict:
              raise TypeError(
                  f'Cannot load class {type_name!r}.\n'
                  'Try pass `auto_dict=True` to load the object into a dict '
                  'without depending on the type.'
              ) from e
        elif not auto_dict:
          raise TypeError(
              f'Type name \'{type_name}\' is not registered '
              'with a `pg.JSONConvertible` subclass.\n'
              'Try pass `auto_import=True` to load the type from its module, '
              'or pass `auto_dict=True` to load the object into a dict '
              'without depending on the type.'
          )

      factory_fn = getattr(cls, 'from_json', None)
      if cls is not None and factory_fn is None and not auto_dict:
        raise TypeError(
            f'{cls} is not a `pg.JSONConvertible` subclass.'
            'Try pass `auto_dict=True` to load the object into a dict '
            'without depending on the type.'
        )

      if factory_fn is None and auto_dict:
        v['type_name'] = type_name
        v.pop(JSONConvertible.TYPE_NAME_KEY)
        return True
      assert factory_fn is not None

    v[JSONConvertible.TYPE_NAME_KEY] = factory_fn
    return True

  def _visit(v) -> None:
    if isinstance(v, (tuple, list)):
      for x in v:
        _visit(x)
    elif isinstance(v, dict):
      if _resolve_typename(v):
        # Only resolve children when _types in this tree is not resolved
        # previously
        for x in getattr(v, 'sym_values', v.values)():
          _visit(x)

  _visit(json_value)
  return json_value


#
# Helper methods for loading/saving Python types and functions.
#


def _type_to_json(t: Type[Any]) -> Dict[str, str]:
  """Converts a type to a JSON dict."""
  type_name = _type_name(t)
  origin = typing.get_origin(t) or t
  if '<locals>' not in type_name and origin is _load_symbol(type_name):
    result = {
        JSONConvertible.TYPE_NAME_KEY: 'type',
        'name': type_name,
    }
    args = typing.get_args(t)
    if args:
      result['args'] = to_json(args)
    return result
  else:
    raise ValueError(f'Cannot convert local class {type_name!r} to JSON.')


def _builtin_function_to_json(f: Any) -> Dict[str, str]:
  return {
      JSONConvertible.TYPE_NAME_KEY: 'function',
      'name': f'builtins.{f.__name__}'
  }


def _function_to_json(f: types.FunctionType) -> Dict[str, str]:
  """Converts a function to a JSON dict."""
  if ('<lambda>' == f.__name__                       # lambda functions.
      or (f.__code__.co_flags & inspect.CO_NESTED)   # local functions.
      ):
    return {
        JSONConvertible.TYPE_NAME_KEY: 'function',
        'name': _type_name(f),
        'code': base64.encodebytes(marshal.dumps(f.__code__)).decode('utf-8'),
        'defaults': to_json(f.__defaults__),
    }

  return {
      JSONConvertible.TYPE_NAME_KEY: 'function',
      'name': _type_name(f)
  }


def _method_to_json(f: types.MethodType) -> Dict[str, str]:
  """Converts a method to a JSON dict."""
  type_name = _type_name(f)
  if isinstance(f.__self__, type):
    return {
        JSONConvertible.TYPE_NAME_KEY: 'method',
        'name': type_name
    }
  raise ValueError(f'Cannot convert instance method {type_name!r} to JSON.')


_SUPPORTED_ANNOTATIONS = {
    typing.Annotated: 'typing.Annotated',
    typing.Any: 'typing.Any',
    typing.Sequence: 'typing.Sequence',
    collections.abc.Sequence: 'typing.Sequence',
    typing.List: 'typing.List',
    list: 'typing.List',
    typing.Tuple: 'typing.Tuple',
    typing.Mapping: 'typing.Mapping',
    collections.abc.Mapping: 'typing.Mapping',
    typing.MutableMapping: 'typing.MutableMapping',
    collections.abc.MutableMapping: 'typing.MutableMapping',
    typing.Dict: 'typing.Dict',
    dict: 'typing.Dict',
    typing.Union: 'typing.Union',
    typing.Optional: 'typing.Optional',
    typing.Callable: 'typing.Callable',
    collections.abc.Callable: 'typing.Callable',
    typing.Set: 'typing.Set',
    set: 'typing.Set',
    typing.FrozenSet: 'typing.FrozenSet',
    frozenset: 'typing.FrozenSet',
}


def _annotation_to_json(annotation) -> Dict[str, str]:
  """Converts a typing annotation to a JSON dict."""
  origin = typing.get_origin(annotation) or annotation
  if origin in _SUPPORTED_ANNOTATIONS:
    name = _SUPPORTED_ANNOTATIONS[origin]
  elif isinstance(origin, type):
    name = _type_name(origin)
  else:
    raise ValueError(f'Annotation cannot be converted to JSON: {annotation}.')

  result = {JSONConvertible.TYPE_NAME_KEY: 'type', 'name': name}
  args = typing.get_args(annotation)
  if args:
    if len(args) > 4:
      raise NotImplementedError(
          'Cannot convert generic type with more than 4 type arguments '
          f'into JSON. Encountered: {annotation}.'
      )
    result['args'] = to_json(args)
  return result


# A symbol cache allows a symbol to be resolved just once upon multiple
# serializations/deserializations.

_LOADED_SYMBOLS = {}

# Special builtin symbols which cannot be accessed from the `builtins` module.

_SPECIAL_BUILTIN_SYMBOLS = {
    'builtins.NoneType': type(None),
    'builtins.Ellipsis': ...,
}


def _load_symbol(type_name: str) -> Any:
  """Loads a symbol from its type name."""
  symbol = _LOADED_SYMBOLS.get(type_name, None)
  if symbol is not None:
    return symbol

  symbol = _SPECIAL_BUILTIN_SYMBOLS.get(type_name, None)
  if symbol is not None:
    _LOADED_SYMBOLS[type_name] = symbol
    return symbol

  # Import symbol based on the module and symbol name.
  *maybe_modules, symbol_name = type_name.split('.')
  module_end_pos = None

  # NOTE(daiyip): symbols could be nested within classes, for example::
  #
  #  class A:
  #    class B:
  #      @classmethod
  #      def x(cls):
  #        pass
  #
  # In such case, class A will have type name `<module>.A`;
  # class B will have type name `module.A.B` and class method `x`
  # will have `module.A.B.x` as its type name.
  #
  # To support nesting, we need to infer the module names from type names
  # correctly. This is done by detecting the first token in the module path
  # whose first letter is capitalized, assuming class names always start
  # with capital letters.
  for i, name in enumerate(maybe_modules):
    if name[0].isupper():
      module_end_pos = i
      break

  # Figure out module path and parent symbol names.
  if module_end_pos is None:
    module_name = '.'.join(maybe_modules)
    parent_symbols = []
  else:
    module_name = '.'.join(maybe_modules[:module_end_pos])
    parent_symbols = maybe_modules[module_end_pos:]

  if not module_name:
    raise ModuleNotFoundError(f'Cannot load symbol {type_name!r}.')

  # Import module and lookup parent symbols.
  module = importlib.import_module(module_name)
  parent = module
  for name in parent_symbols:
    parent = getattr(parent, name)

  # Lookup the final symbol.
  symbol = getattr(parent, symbol_name)
  _LOADED_SYMBOLS[type_name] = symbol
  return symbol


def _type_from_json(json_value: Dict[str, str], **kwargs) -> Type[Any]:
  """Loads a type from a JSON dict."""
  del kwargs
  t = _load_symbol(json_value['name'])
  if 'args' in json_value:
    return _bind_type_args(
        t, from_json(json_value['args'], _typename_resolved=True)
    )
  return t


def _function_from_json(
    json_value: Dict[str, str], **kwargs) -> types.FunctionType:
  """Loads a function from a JSON dict."""
  del kwargs
  function_name = json_value['name']
  if 'code' in json_value:
    code = marshal.loads(
        base64.decodebytes(json_value['code'].encode('utf-8')))
    defaults = from_json(json_value['defaults'], _typename_resolved=True)
    return types.FunctionType(
        code=code,
        globals=globals(),
        argdefs=defaults,
    )
  else:
    return _load_symbol(function_name)


def _method_from_json(json_value: Dict[str, str], **kwargs) -> types.MethodType:
  """Loads a class method from a JSON dict."""
  del kwargs
  return _load_symbol(json_value['name'])


def _bind_type_args(t, args):
  """Bind type args to a type."""
  # NOTE(daiyip): Haven't found an equivalence for expressing `t[**args]`,
  # thus we hard code the logic based on the number of type args. May change
  # if future when we find better ways of expressing this.
  assert args and len(args) <= 4, args
  if len(args) == 1:
    return t[args[0]]
  elif len(args) == 2:
    return t[args[0], args[1]]
  elif len(args) == 3:
    return t[args[0], args[1], args[2]]
  else:
    return t[args[0], args[1], args[2], args[3]]



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom hyper primitives."""

import abc
import random
import types
from typing import Any, Callable, Optional, Tuple, Union

from pyglove.core import geno
from pyglove.core import typing as pg_typing
from pyglove.core import utils
from pyglove.core.hyper import base


class CustomHyper(base.HyperPrimitive):
  """User-defined hyper primitive.

  User-defined hyper primitive is useful when users want to have full control
  on the semantics and genome encoding of the search space. For example, the
  decision points are of variable length, which is not yet supported by
  built-in hyper primitives.

  To use user-defined hyper primitive is simple, the user should:


  1) Subclass `CustomHyper` and implements the
     :meth:`pyglove.hyper.CustomHyper.custom_decode` method.
     It's optional to implement the
     :meth:`pyglove.hyper.CustomHyper.custom_encode` method, which is only
     necessary when the user want to encoder a material object into a DNA.
  2) Introduce a DNAGenerator that can generate DNA for the
     :class:`pyglove.geno.CustomDecisionPoint`.

  For example, the following code tries to find an optimal sub-sequence of an
  integer sequence by their sums::

    import random

    class IntSequence(pg.hyper.CustomHyper):

      def custom_decode(self, dna):
        return [int(v) for v in dna.value.split(',') if v != '']

    class SubSequence(pg.evolution.Mutator):

      def mutate(self, dna):
        genome = dna.value
        items = genome.split(',')
        start = random.randint(0, len(items))
        end = random.randint(start, len(items))
        new_genome = ','.join(items[start:end])
        return pg.DNA(new_genome, spec=dna.spec)

    @pg.geno.dna_generator
    def initial_population():
      yield pg.DNA('12,-34,56,-2,100,98', spec=dna_spec)

    algo = pg.evolution.Evolution(
        (pg.evolution.selectors.Random(10)
         >> pg.evolution.selectors.Top(1)
         >> SubSequence()),
        population_init=initial_population(),
        population_update=pg.evolution.selectors.Last(20))

    best_reward, best_example = None, None
    for int_seq, feedback in pg.sample(IntSequence(), algo, num_examples=100):
      reward = sum(int_seq)
      if best_reward is None or best_reward < reward:
        best_reward, best_example = reward, int_seq
      feedback(reward)

    print(best_reward, best_example)

  Please note that user-defined hyper value can be used together with PyGlove's
  built-in hyper primitives, for example::

    pg.oneof([IntSequence(), None])

  Therefore it's also a mechanism to extend PyGlove's search space definitions.
  """

  def _decode(self):
    if not isinstance(self.dna.value, str):
      raise ValueError(
          f'{self.__class__} expects string type DNA. '
          f'Encountered {self.dna!r}.')
    return self.custom_decode(self.dna)

  @abc.abstractmethod
  def custom_decode(self, dna: geno.DNA) -> Any:
    """Decode a DNA whose value is a string of user-defined genome."""

  def encode(self, value: Any) -> geno.DNA:
    """Encode value into DNA with user-defined genome."""
    return self.custom_encode(value)

  def custom_encode(self, value: Any) -> geno.DNA:
    """Encode value to user defined genome."""
    raise NotImplementedError(
        f'\'custom_encode\' is not supported by {self.__class__.__name__!r}.')

  def dna_spec(self, location: Optional[utils.KeyPath] = None) -> geno.DNASpec:
    """Always returns CustomDecisionPoint for CustomHyper."""
    return geno.CustomDecisionPoint(
        hyper_type=self.__class__.__name__,
        next_dna_fn=self.next_dna,
        random_dna_fn=self.random_dna,
        hints=self.hints, name=self.name, location=location)

  def first_dna(self) -> geno.DNA:
    """Returns the first DNA of current sub-space.

    Returns:
      A string-valued DNA.
    """
    if self.next_dna.__code__ is CustomHyper.next_dna.__code__:
      raise NotImplementedError(
          f'{self.__class__!r} must implement method `next_dna` to be used in '
          f'dynamic evaluation mode.')
    return self.next_dna(None)

  def next_dna(self, dna: Optional[geno.DNA] = None) -> Optional[geno.DNA]:
    """Subclasses should override this method to support pg.Sweeping."""
    raise NotImplementedError(
        f'`next_dna` is not implemented in f{self.__class__!r}')

  def random_dna(
      self,
      random_generator: Union[types.ModuleType, random.Random, None] = None,
      previous_dna: Optional[geno.DNA] = None) -> geno.DNA:
    """Subclasses should override this method to support pg.random_dna."""
    raise NotImplementedError(
        f'`random_dna` is not implemented in {self.__class__!r}')

  def custom_apply(
      self,
      path: utils.KeyPath,
      value_spec: pg_typing.ValueSpec,
      allow_partial: bool,
      child_transform: Optional[
          Callable[[utils.KeyPath, pg_typing.Field, Any], Any]
      ] = None,
  ) -> Tuple[bool, 'CustomHyper']:
    """Validate candidates during value_spec binding time."""
    del path, value_spec, allow_partial, child_transform
    # Allow custom hyper to be assigned to any type.
    return (False, self)



# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import tempfile
import unittest
from pyglove.core.io import file_system


class StdFileSystemTest(unittest.TestCase):

  def test_file(self):
    tmp_dir = tempfile.mkdtemp()
    fs = file_system.StdFileSystem()

    file1 = os.path.join(tmp_dir, 'file1')
    with self.assertRaises(FileNotFoundError):
      fs.open(file1)

    with fs.open(file1, 'w') as f:
      f.write('hello\npyglove')

    with fs.open(file1, 'r') as f:
      self.assertEqual(f.readline(), 'hello\n')
      self.assertEqual(f.tell(), 6)
      self.assertEqual(f.seek(8, 0), 8)
      self.assertEqual(f.read(), 'glove')

    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.exists('/not-exist-dir'))
    fs.rm(file1)
    self.assertFalse(fs.exists(file1))

  def test_file_system(self):
    tmp_dir = tempfile.mkdtemp()
    fs = file_system.StdFileSystem()

    # Create a directory.
    dir_a = os.path.join(tmp_dir, 'a')
    fs.mkdir(dir_a)
    self.assertTrue(fs.exists(dir_a))
    self.assertTrue(fs.isdir(dir_a))
    self.assertEqual(len(fs.listdir(dir_a)), 0)

    # Create a file under the directory.
    file1 = os.path.join(dir_a, 'file1')
    self.assertFalse(fs.exists(file1))
    with fs.open(file1, 'w') as f:
      f.write('hello')
    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.isdir(file1))

    fs.mkdirs(os.path.join(dir_a, 'b/c/d'))
    self.assertEqual(sorted(fs.listdir(dir_a)), ['b', 'file1'])   # pylint: disable=g-generic-assert

    # Test rm.
    with self.assertRaises(FileNotFoundError):
      fs.rm(os.path.join(dir_a, 'file2'))
    with self.assertRaises((IsADirectoryError, PermissionError)):
      fs.rm(os.path.join(dir_a, 'b'))

    # Test rmdir.
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'c'))
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'file2'))
    fs.rmdir(os.path.join(dir_a, 'b/c/d'))

    # Test rmdirs.
    fs.rmdirs(os.path.join(dir_a, 'b/c'))
    self.assertEqual(sorted(fs.listdir(dir_a)), ['file1'])   # pylint: disable=g-generic-assert


class MemoryFileSystemTest(unittest.TestCase):

  def test_file(self):
    fs = file_system.MemoryFileSystem()

    file1 = os.path.join('/mem', 'file1')
    with self.assertRaises(FileNotFoundError):
      fs.open(file1)

    with self.assertRaisesRegex(ValueError, 'Unsupported path'):
      fs.open(1)

    with fs.open(file1, 'w') as f:
      f.write('hello\npyglove')

    with fs.open(file1, 'r') as f:
      self.assertEqual(f.readline(), 'hello\n')
      self.assertEqual(f.tell(), 6)
      self.assertEqual(f.seek(8, 0), 8)
      self.assertEqual(f.read(), 'glove')

    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.exists('/not-exist-dir'))
    fs.rm(file1)
    self.assertFalse(fs.exists(file1))

  def test_file_system(self):
    fs = file_system.MemoryFileSystem()

    # Create a directory.
    dir_a = os.path.join('/mem', 'a')
    fs.mkdir(dir_a)
    self.assertTrue(fs.exists(dir_a))
    self.assertTrue(fs.isdir(dir_a))
    self.assertEqual(len(fs.listdir(dir_a)), 0)

    with self.assertRaises(FileExistsError):
      fs.mkdir(dir_a)

    with self.assertRaises(FileNotFoundError):
      fs.mkdir('/mem/b/c')

    with self.assertRaises(IsADirectoryError):
      fs.open(dir_a)

    # Create a file under the directory.
    file1 = os.path.join(dir_a, 'file1')
    self.assertFalse(fs.exists(file1))
    with fs.open(file1, 'w') as f:
      f.write('hello')
      f.flush()
    self.assertTrue(fs.exists(file1))
    self.assertFalse(fs.isdir(file1))

    # Make dirs.
    fs.mkdirs(os.path.join(dir_a, 'b/c/d'))

    fs.mkdirs(os.path.join(dir_a, 'b/c/d'))
    with self.assertRaises(FileExistsError):
      fs.mkdirs(os.path.join(dir_a, 'b/c/d'), exist_ok=False)
    with self.assertRaises(NotADirectoryError):
      fs.mkdirs(os.path.join(file1, 'e'))

    with self.assertRaises(FileNotFoundError):
      fs.listdir('/mem/not-exist')
    self.assertEqual(sorted(fs.listdir(dir_a)), ['b', 'file1'])   # pylint: disable=g-generic-assert

    # Test rm.
    with self.assertRaises(FileNotFoundError):
      fs.rm(os.path.join(dir_a, 'file2'))
    with self.assertRaises(IsADirectoryError):
      fs.rm(os.path.join(dir_a, 'b'))

    # Test rmdir.
    with self.assertRaisesRegex(OSError, 'Directory not empty'):
      fs.rmdir(dir_a)
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'c'))
    with self.assertRaises(FileNotFoundError):
      fs.rmdir(os.path.join(dir_a, 'file2'))
    with self.assertRaises(NotADirectoryError):
      fs.rmdir(os.path.join(dir_a, 'file1'))
    fs.rmdir(os.path.join(dir_a, 'b/c/d'))

    # Test rmdirs.
    with self.assertRaisesRegex(OSError, 'Directory not empty'):
      fs.rmdirs(dir_a)

    with self.assertRaises(NotADirectoryError):
      fs.rmdirs(os.path.join(file1, 'a'))

    with self.assertRaises(FileNotFoundError):
      fs.rmdirs('a/b/d')

    fs.rmdirs(os.path.join(dir_a, 'b/c'))
    self.assertEqual(fs.listdir(dir_a), ['file1'])   # pylint: disable=g-generic-assert


class FileIoApiTest(unittest.TestCase):

  def test_standard_filesystem(self):
    file1 = os.path.join(tempfile.mkdtemp(), 'file1')
    with self.assertRaises(FileNotFoundError):
      file_system.readfile(file1)
    self.assertIsNone(file_system.readfile(file1, nonexist_ok=True))
    with file_system.open(file1, 'w') as f:
      self.assertIsInstance(f, file_system.StdFile)
      f.write('foo')
    self.assertTrue(file_system.path_exists(file1))
    self.assertEqual(file_system.readfile(file1), 'foo')
    file_system.writefile(file1, 'bar')
    self.assertEqual(file_system.readfile(file1), 'bar')
    self.assertFalse(file_system.isdir(file1))
    file_system.rm(file1)
    self.assertFalse(file_system.path_exists(file1))

    dir1 = os.path.join(tempfile.mkdtemp(), 'dir1')
    file_system.mkdir(dir1)
    self.assertEqual(file_system.listdir(dir1), [])
    file_system.mkdirs(os.path.join(dir1, 'a/b/c'))
    file2 = os.path.join(dir1, 'file2')
    file_system.writefile(file2, 'baz')
    self.assertEqual(sorted(file_system.listdir(dir1)), ['a', 'file2'])  # pylint: disable=g-generic-assert
    self.assertEqual(    # pylint: disable=g-generic-assert
        sorted(file_system.listdir(dir1, fullpath=True)),
        [os.path.join(dir1, 'a'), os.path.join(dir1, 'file2')]
    )
    file_system.rmdir(os.path.join(dir1, 'a/b/c'))
    file_system.rmdirs(os.path.join(dir1, 'a/b'))
    self.assertTrue(file_system.path_exists(dir1))
    self.assertEqual(sorted(file_system.listdir(dir1)), ['file2'])  # pylint: disable=g-generic-assert
    file_system.rm(file2)
    self.assertFalse(file_system.path_exists(file2))

  def test_memory_filesystem(self):
    file1 = pathlib.Path('/mem/file1')
    with self.assertRaises(FileNotFoundError):
      file_system.readfile(file1)
    self.assertIsNone(file_system.readfile(file1, nonexist_ok=True))
    with file_system.open(file1, 'w') as f:
      self.assertIsInstance(f, file_system.MemoryFile)
      f.write('foo')
      f.flush()
    self.assertTrue(file_system.path_exists(file1))
    self.assertEqual(file_system.readfile(file1), 'foo')
    file_system.writefile(file1, 'bar')
    self.assertEqual(file_system.readfile(file1), 'bar')
    self.assertFalse(file_system.isdir(file1))
    file_system.rm(file1)
    self.assertFalse(file_system.path_exists(file1))

    dir1 = os.path.join('/mem/dir1')
    file_system.mkdir(dir1)
    self.assertEqual(file_system.listdir(dir1), [])
    file_system.mkdirs(os.path.join(dir1, 'a/b/c'))
    file2 = os.path.join(dir1, 'file2')
    file_system.writefile(file2, 'baz')
    self.assertEqual(sorted(file_system.listdir(dir1)), ['a', 'file2'])  # pylint: disable=g-generic-assert
    file_system.rmdir(os.path.join(dir1, 'a/b/c'))
    file_system.rmdirs(os.path.join(dir1, 'a/b'))
    self.assertTrue(file_system.path_exists(dir1))
    self.assertEqual(sorted(file_system.listdir(dir1)), ['file2'])  # pylint: disable=g-generic-assert
    file_system.rm(file2)
    self.assertFalse(file_system.path_exists(file2))


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from pyglove.core import symbolic
from pyglove.core import utils
from pyglove.core.geno.base import ConditionalKey
from pyglove.core.geno.base import DNA
from pyglove.core.geno.categorical import manyof
from pyglove.core.geno.categorical import oneof
from pyglove.core.geno.custom import custom
from pyglove.core.geno.numerical import floatv
from pyglove.core.geno.space import constant
from pyglove.core.geno.space import Space


class DNATest(unittest.TestCase):
  """Tests for `pg.DNA`."""

  def _dna(self):
    spec = Space([
        oneof([
            manyof(2, [
                custom(),
                floatv(0.0, 1.0, name='z'),
                constant(),
            ], location='b', name='y'),
            constant()
        ], literal_values=[
            '0/2 (Something complex)', '1/2 (\'foo\')'
        ], location='a', name='x'),
        floatv(-1.0, 0.0, location='d', name='p'),
        # NOTE(daiyip): multiple choices NOT under conditions.
        manyof(2, [
            constant(), constant(), constant()
        ], literal_values=[
            '0/3 (\'a\')',
            '1/3 (\'b\')',
            '2/3 (\'c\')',
        ], location='e', name='q')
    ])
    return DNA([(0, [(0, 'abc'), (1, 0.5)]), -0.5, [0, 1]]).use_spec(spec)

  def test_init(self):
    dna_spec = floatv(min_value=0.0, max_value=1.0)
    dna = DNA(0.5, spec=dna_spec)
    self.assertEqual(dna.spec, dna_spec)

    # Reduant None-value ancestors. (single node)
    # should be reduced to the DNA value itself.
    self.assertEqual(
        DNA(None, [DNA(None, [DNA(None, [DNA(1)])])]), DNA(1))

    # Reduant None-value ancestors. (multiple nodes)
    # should be reduced to 2-level DNA: a None parent and children with values.
    self.assertEqual(
        DNA(None, [DNA(None, [DNA(None, [DNA(1), DNA('abc')])])]),
        DNA(None, [DNA(1), DNA('abc')]))

    # No redundant node in the DNA tree, remains the same.
    self.assertEqual(
        DNA(None, [DNA(1), DNA(None, [DNA(2), DNA(None, [DNA(3)])])]),
        DNA(None, [DNA(1), DNA(None, [DNA(2), DNA(None, [DNA(3)])])]))

    # Tests for compositional values.
    self.assertEqual(DNA([1, 2]), DNA(None, [DNA(1), DNA(2)]))
    self.assertEqual(DNA((1, 1)), DNA(1, [DNA(1)]))
    self.assertEqual(DNA((1, 2, [3, 4])), DNA(1, [DNA(2, [DNA(3), DNA(4)])]))

    self.assertEqual(
        DNA([(1, [2, (3, 4)]), 5, (6, [7, 8])]),
        DNA(None, [
            DNA(1, [DNA(2), DNA(3, [DNA(4)])]),
            DNA(5),
            DNA(6, [DNA(7), DNA(8)])
        ]))

  def test_bad_init(self):
    with self.assertRaisesRegex(TypeError, 'Expect .* but encountered .*'):
      DNA(ValueError())

    with self.assertRaisesRegex(
        ValueError,
        '\'children\' .* must be None when \'value\' .* is compositional.'):
      DNA([1, 2], [DNA(1)])

    with self.assertRaisesRegex(
        ValueError,
        'Tuple as conditional choices must have at least 2 items'):
      DNA((1,))

    with self.assertRaisesRegex(
        ValueError, 'Tuple as conditional choices only allow multiple '
        'choices to be used at the last position'):
      DNA([(1, [1, 2], 3)])

  def test_basics(self):
    dna = DNA(1, [DNA(2), DNA(3.0), DNA('abc')])
    self.assertEqual(dna.value, 1)
    self.assertIsNotNone(dna.children)
    self.assertEqual(len(dna.children), 3)
    self.assertIsNone(dna.spec)
    self.assertIsNone(dna.parent_dna)
    self.assertIs(dna.children[0].parent_dna, dna)
    self.assertIs(dna.root, dna)
    self.assertIs(dna.children[0].root, dna)

  def test_contains(self):
    dna = DNA(1, [DNA(2), DNA(3.0), DNA('abc')])
    self.assertIn(2, dna)
    self.assertIn(3.0, dna)
    self.assertIn('abc', dna)
    self.assertIn(DNA(2), dna)
    self.assertNotIn(0.1, dna)
    self.assertNotIn('foo', dna)
    with self.assertRaisesRegex(
        ValueError, 'DNA.__contains__ does not accept .*'):
      _ = (1, 1) in dna

  def test_iter(self):
    dna = DNA(1, [DNA(2), DNA(3.0), DNA('abc')])
    values = []
    for child in dna:
      values.append(child.value)
    self.assertEqual(values, [2, 3.0, 'abc'])

  def test_is_leaf(self):
    dna = DNA(1, [DNA(2), DNA(3.0), DNA('abc')])
    self.assertFalse(dna.is_leaf)
    self.assertTrue(dna.children[0].is_leaf)
    self.assertTrue(dna.children[1].is_leaf)
    self.assertTrue(dna.children[2].is_leaf)

  def test_methods_that_require_spec(self):
    dna = DNA(1, [DNA(2), DNA(3.0), DNA('abc')])
    with self.assertRaisesRegex(ValueError, '.* is not bound with a DNASpec'):
      dna.get('a')

    with self.assertRaisesRegex(ValueError, '.* is not bound with a DNASpec'):
      _ = dna.literal_value

    with self.assertRaisesRegex(ValueError, '.* is not bound with a DNASpec'):
      _ = dna.decision_ids

    with self.assertRaisesRegex(ValueError, '.* is not bound with a DNASpec'):
      _ = dna.to_dict()

  def test_inspection(self):
    self.assertEqual(DNA(0).to_json(type_info=False), 0)
    self.assertEqual(DNA([0, 1]).to_json(type_info=False), [0, 1])
    self.assertEqual(
        DNA([0, (0, 1)]).to_json(type_info=False), [0, (0, 1)])
    self.assertEqual(
        DNA((0, 0.5)).to_json(type_info=False), (0, 0.5))

    self.assertEqual(
        str(DNA((0, [1, 0.5, 'abc']))), 'DNA(0, [1, 0.5, \'abc\'])')
    self.assertEqual(
        DNA(0.0).use_spec(
            Space([
                floatv(0.0, 1.0, location='a')
            ])).format(as_dict=True), 'DNA({\n'
        '  \'a\': 0.0\n'
        '})')

  def test_from_fn(self):
    dna_spec = Space([
        oneof([
            manyof(2, [constant(), constant()]),
            constant()]),
        floatv(0.0, 1.0)
    ])

    stats = dict(invocation=0)
    def dna_emitter(spec):
      stats['invocation'] += 1
      return spec.first_dna()

    # `dna_emitter` should be called only twice, once on
    # the top-most Choices and once on the Float. Since both
    # of them returns a DNA, we use its child DNA directly
    # instead of stepping into the child DNASpec.
    dna = DNA.from_fn(dna_spec, dna_emitter)
    self.assertEqual(stats['invocation'], 2)
    self.assertEqual(dna, DNA([(0, [0, 1]), 0.0]))

    def zero_emitter(spec):
      if spec.is_numerical:
        return spec.min_value
      return range(spec.num_choices)

    dna = DNA.from_fn(dna_spec, zero_emitter)
    self.assertEqual(dna, DNA([(0, [0, 1]), 0.0]))

    def choice_out_of_range(spec):
      if spec.is_numerical:
        return spec.min_value
      return [len(spec.candidates)] * spec.num_choices

    with self.assertRaisesRegex(ValueError, 'Choice out of range.'):
      DNA.from_fn(dna_spec, choice_out_of_range)

    def bad_choice_value(spec):
      del spec
      return [0.1]

    with self.assertRaisesRegex(ValueError, 'Choice value should be int'):
      DNA.from_fn(dna_spec, bad_choice_value)

    def unmatched_num_choices(spec):
      del spec
      return [0]

    with self.assertRaisesRegex(
        ValueError,
        'Number of DNA child values does not match the number of choices'):
      DNA.from_fn(dna_spec, unmatched_num_choices)

    with self.assertRaisesRegex(
        TypeError, 'Argument \'dna_spec\' should be DNASpec type.'):
      DNA.from_fn(1, unmatched_num_choices)

  def test_serialization(self):
    dna = DNA([0, (1, 2), [3, (4, 5, 'abc')]])
    dna.set_metadata('a', 1, cloneable=True)
    dna_str = symbolic.to_json_str(dna)
    dna2 = symbolic.from_json_str(dna_str)
    self.assertEqual(dna, dna2)
    self.assertEqual(dna.metadata, dna2.metadata)

    # Serialization using non-compact form.
    dna_str_verbose = symbolic.to_json_str(dna, compact=False)
    self.assertGreater(len(dna_str_verbose), len(dna_str))
    dna3 = symbolic.from_json_str(dna_str_verbose)
    self.assertEqual(dna, dna3)
    self.assertEqual(dna.metadata, dna3.metadata)

  def test_decision_ids(self):
    self.assertEqual(
        self._dna().decision_ids,
        [
            'a',
            'a[=0/2].b',
            'a[=0/2].b[0][=0/3]',
            'a[=0/2].b[0][=1/3]',
            'a[=0/2].b[1][=0/3]',
            'a[=0/2].b[1][=1/3]',
            'd',
            'e'
        ])

  def test_getitem(self):
    # Get by index.
    dna = DNA(1, [DNA(2), DNA(3.0), DNA('abc')])
    self.assertEqual(dna[0], DNA(2))
    with self.assertRaisesRegex(IndexError, 'list index out of range'):
      _ = dna[3]

    # Get by name.
    dna = self._dna()
    self.assertEqual(dna['x'], DNA((0, [(0, 'abc'), (1, 0.5)])))
    self.assertEqual(dna['y'], [DNA((0, 'abc')), DNA((1, 0.5))])
    self.assertEqual(dna['z'], DNA(0.5))
    self.assertEqual(dna['p'], DNA(-0.5))
    self.assertEqual(dna['q'], [DNA(0), DNA(1)])

    # Get by id (Text).
    self.assertEqual(dna['a'], DNA((0, [(0, 'abc'), (1, 0.5)])))
    self.assertEqual(dna['a[=0/2].b'], [DNA((0, 'abc')),
                                        DNA((1, 0.5))])
    self.assertEqual(dna['a[=0/2].b[0]'], DNA((0, 'abc')))
    self.assertEqual(dna['a[=0/2].b[0][=0/3]'], DNA('abc'))
    self.assertIsNone(dna['a[=0/2].b[0][=1/3]'])
    self.assertIsNone(dna['a[=0/2].b[1][=0/3]'])
    self.assertEqual(dna['a[=0/2].b[1]'], DNA((1, 0.5)))
    self.assertEqual(dna['a[=0/2].b[1][=1/3]'], DNA(0.5))
    self.assertEqual(dna['d'], DNA(-0.5))
    self.assertEqual(dna['e'], [DNA(0), DNA(1)])
    self.assertEqual(dna['e[0]'], DNA(0))

    # Get by id (KeyPath).
    self.assertEqual(dna[dna.spec['a[=0/2].b[1][=1/3]'].id], DNA(0.5))

    # Get by DNASpec.
    self.assertEqual(dna[dna.spec['a[=0/2].b[1][=1/3]']], DNA(0.5))

  def test_get(self):
    dna = self._dna()
    self.assertEqual(dna.get('a[=0/2].b'), [DNA((0, 'abc')),
                                            DNA((1, 0.5))])
    self.assertEqual(dna.get('a[=0/2].b[1]'), DNA((1, 0.5)))
    self.assertIsNone(dna.get('a[=0/2].b[0][=1/3]'))
    self.assertIsNone(dna.get('xyz'))

  def test_literal_value(self):
    dna = self._dna()
    self.assertEqual(
        dna.literal_value,
        [
            ['0/2 -> 0/3 -> abc', '0/2 -> 1/3 -> 0.5'],
            '-0.5',
            ["'a'", "'b'"]
        ])
    self.assertEqual(dna['y'][0].literal_value, '0/3 -> abc')
    self.assertEqual(dna['z'].literal_value, '0.5')
    self.assertEqual(dna['q'][0].literal_value, '\'a\'')

  def test_is_subchoice(self):
    dna = self._dna()
    self.assertFalse(dna.is_subchoice)
    self.assertTrue(dna['y'][0].is_subchoice)
    self.assertFalse(dna['z'].is_subchoice)
    self.assertTrue(dna['q'][1].is_subchoice)

  def test_root(self):
    dna = self._dna()
    self.assertIs(dna['y'][0].root, dna)
    self.assertIs(dna['z'].root, dna)
    self.assertIs(dna['q'][0].root, dna)

  def test_hash(self):
    self.assertEqual(hash(DNA(None)), hash(DNA(None)))
    self.assertEqual(hash(DNA(1)), hash(DNA(1)))
    self.assertEqual(hash(DNA(1, [DNA(2, [DNA(3), DNA(4)]), DNA(5)])),
                     hash(DNA(1, [DNA(2, [DNA(3), DNA(4)]), DNA(5)])))

  def test_cmp(self):
    self.assertLess(DNA(None), DNA(0))
    self.assertLess(DNA(0), DNA(1))
    self.assertLess(DNA([0, 1, 2]), DNA([0, 2, 1]))
    self.assertLess(DNA((0, 'abc')), DNA((0, 'abd')))
    self.assertLess(DNA((0, 1)), DNA((0, 'abd')))
    self.assertLess(DNA([0, 1]), DNA(0))
    self.assertLess(DNA((0, 0, 0)), DNA((0, 0, 1, 0)))

    self.assertGreater(DNA(0), DNA(None))
    self.assertGreater(DNA(1), DNA(0))
    self.assertGreater(DNA([0, 2, 1]), DNA([0, 1, 2]))
    self.assertGreater(DNA((0, 'abd')), DNA((0, 'abc')))
    self.assertGreater(DNA((0, 'abd')), DNA((0, 1)))
    self.assertGreater(DNA(0), DNA([0, 1]))
    self.assertGreater(DNA((0, 0, 1, 0)), DNA((0, 0, 0)))

    with self.assertRaisesRegex(
        ValueError,
        'The two input DNA have different number of children'):
      _ = DNA((0, 0, 0)) < DNA((0, 0, 0, 0))

  def test_to_numbers(self):
    self.assertEqual(DNA(None).to_numbers(), [])
    self.assertEqual(DNA(None).to_numbers(flatten=False), [])
    self.assertEqual(DNA(1).to_numbers(), [1])
    self.assertEqual(DNA(1).to_numbers(flatten=False), 1)
    self.assertEqual(DNA(0.5).to_numbers(), [0.5])
    self.assertEqual(DNA(0.5).to_numbers(flatten=False), 0.5)
    self.assertEqual(DNA([0, 1, 2]).to_numbers(), [0, 1, 2])
    self.assertEqual(
        DNA((0, [1, 2])).to_numbers(flatten=False),
        (0, [1, 2]))
    self.assertEqual(
        DNA([0, 1, 2]).to_numbers(flatten=False),
        [0, 1, 2])
    self.assertEqual(
        DNA([(0, 1), 2, (3, [4, 0.5])]).to_numbers(),
        [0, 1, 2, 3, 4, 0.5])
    self.assertEqual(
        DNA([(0, 1), 'abc', (3, [4, 0.5])]).to_numbers(flatten=False),
        [(0, 1), 'abc', (3, [4, 0.5])])

  def test_from_numbers(self):
    spec = Space([
        manyof(2, [
            oneof([
                floatv(0.0, 1.0),
                constant()
            ]),
            constant(),
        ]),
        floatv(-1., 1.)
    ])
    self.assertEqual(
        DNA.from_numbers([0, 0, 0.1, 1, 0.2], spec),
        DNA([[(0, 0, 0.1), 1], 0.2]))

    with self.assertRaisesRegex(
        ValueError, 'Candidate index out of range at choice .*'):
      DNA.from_numbers([0, 4, 0.1, 1, 0.2], spec)

    with self.assertRaisesRegex(
        ValueError, 'The input .* is too short'):
      DNA.from_numbers([0, 0], spec)

    with self.assertRaisesRegex(
        ValueError, 'The input .* is too long'):
      DNA.from_numbers([0, 0, 0.1, 1, 0.2, 0], spec)

  def test_metadata(self):
    dna = DNA(None)
    self.assertEqual(len(dna.metadata), 0)
    self.assertEqual(dna.set_metadata('a', 1).metadata.a, 1)

    dna = DNA(None, metadata=dict(a=1))
    self.assertEqual(dna.metadata, dict(a=1))
    self.assertEqual(dna.set_metadata('a', 2).set_metadata('b', 3).metadata,
                     dict(a=2, b=3))
    # Neither metadata keys are cloneable.
    self.assertEqual(dna.clone(deep=True).metadata, {})

    dna_str = symbolic.to_json_str(dna)
    dna2 = symbolic.from_json_str(dna_str)
    self.assertEqual(dna2.metadata, dict(a=2, b=3))

    dna = DNA(None, metadata=dict(a=1))
    dna.set_metadata('b', 2, cloneable=True)
    dna.set_metadata('c', 3, cloneable=False)
    self.assertEqual(dna._cloneable_metadata_keys, set('b'))    # pylint: disable=protected-access
    # Only metadata 'b' is cloneable.
    self.assertEqual(dna.clone(deep=True).metadata, dict(b=2))

    json_value = symbolic.to_json(dna)
    self.assertEqual(json_value['_cloneable_metadata_keys'], ['b'])
    dna2 = symbolic.from_json(json_value)
    # All metadata should be present during serialization but only cloneable
    # keys should be present after cloning.
    self.assertEqual(dna2.metadata, dict(a=1, b=2, c=3))
    self.assertEqual(dna2._cloneable_metadata_keys, set('b'))    # pylint: disable=protected-access
    self.assertEqual(dna2.clone(deep=True).metadata, dict(b=2))

  def test_userdata(self):
    dna = DNA(None)
    self.assertEqual(len(dna.userdata), 0)
    self.assertEqual(dna.set_userdata('a', 1).userdata.a, 1)
    self.assertEqual(dna.set_userdata('b', 'foo', True).userdata.b, 'foo')

    # Test clone carries over the userdata 'b'
    dna2 = dna.clone()
    self.assertNotIn('a', dna2.userdata)
    self.assertEqual(dna2.userdata.b, 'foo')

  def test_iter_dna(self):
    dna_spec = oneof([
        constant(), constant(), constant(),
        constant(), constant(), constant()])
    dna = DNA(3, spec=dna_spec)
    self.assertEqual(dna.next_dna(), DNA(4))
    self.assertEqual(list(dna.iter_dna()), [DNA(4), DNA(5)])

    with self.assertRaisesRegex(
        ValueError, '.* is not bound with a DNASpec'):
      DNA(3).next_dna()

    with self.assertRaisesRegex(
        ValueError, '.* is not bound with a DNASpec'):
      next(DNA(3).iter_dna())

  def test_named_decisions(self):
    t = Space([
        manyof(3, [
            Space([
                manyof(2, [
                    floatv(0.1, 0.9, name='c', location='z'),
                    oneof([
                        constant(),
                        constant(),
                        constant(),
                    ], name='d'),
                    constant(),
                    constant(),
                ], location='y', name='b'),
                oneof([
                    constant(),
                    constant()
                ], location='q')
            ]),
            constant(),
            constant()
        ], distinct=False, location='x', name='a'),
        floatv(location='m', name='e', min_value=0.0, max_value=1.0)
    ])
    dna = DNA([
        [(0, [[(0, 0.1), (1, 0)], 0]),
         (0, [[(0, 0.2), (1, 1)], 0]),
         (0, [[(0, 0.3), (1, 2)], 0])],
        0.0
    ], spec=t)
    self.assertEqual(
        list(dna.named_decisions.keys()), ['a', 'b', 'c', 'd', 'e'])
    self.assertEqual(dna.named_decisions['a'], [
        DNA((0, [[(0, 0.1), (1, 0)], 0])),
        DNA((0, [[(0, 0.2), (1, 1)], 0])),
        DNA((0, [[(0, 0.3), (1, 2)], 0])),
    ])
    self.assertEqual(dna.named_decisions['b'], [
        DNA((0, 0.1)),
        DNA((1, 0)),
        DNA((0, 0.2)),
        DNA((1, 1)),
        DNA((0, 0.3)),
        DNA((1, 2))
    ])
    self.assertEqual(dna.named_decisions['c'], [
        DNA(0.1), None, DNA(0.2), None, DNA(0.3), None
    ])
    self.assertEqual(dna.named_decisions['d'], [
        DNA(0), None, DNA(1), None, DNA(2)
    ])
    self.assertEqual([d.parent_dna for d in dna.named_decisions['d'] if d], [
        DNA((1, 0)),
        DNA((1, 1)),
        DNA((1, 2))
    ])
    self.assertEqual(dna.named_decisions['e'], DNA(0.0))

  def test_multi_choice_spec(self):
    dna_spec = manyof(2, [
        Space([
            oneof([constant(), constant()]),
            oneof([constant(), constant()])
        ]),
        manyof(2, [
            constant(),
            constant()
        ])
    ])
    dna = DNA([(0, [1, 1]), (1, [0, 1])], spec=dna_spec)
    self.assertIs(dna.multi_choice_spec, dna_spec)
    self.assertTrue(dna.is_multi_choice_container)

    self.assertIsNone(dna.children[0].multi_choice_spec)
    self.assertFalse(dna.children[0].is_multi_choice_container)

    self.assertIsNone(dna.children[0].children[0].multi_choice_spec)
    self.assertFalse(dna.children[0].children[0].is_multi_choice_container)

    self.assertIsNotNone(dna.children[1].multi_choice_spec)
    self.assertTrue(dna.children[1].is_multi_choice_container)

    self.assertIsNone(dna.children[1].children[0].multi_choice_spec)
    self.assertFalse(dna.children[1].children[0].is_multi_choice_container)


class DNAUseSpecTest(unittest.TestCase):
  """Tests for `pg.DNA.use_spec`."""

  def test_use_spec_on_oneof(self):
    spec = oneof([
        constant(),
        Space([
            floatv(0.0, 1.0, hints=0),
            floatv(0.0, 1.0),
            custom()
        ]),
        Space([floatv(0.0, 1.0, hints=2)]),
    ], hints=1)
    self.assertEqual(DNA(0).use_spec(spec).spec, spec)

    dna = DNA((1, [0.0, 1.0, 'abc'])).use_spec(spec)
    self.assertIs(dna.spec, spec)
    self.assertTrue(symbolic.eq(
        dna.children[0].spec,
        floatv(min_value=0.0, max_value=1.0, hints=0)))
    self.assertTrue(symbolic.eq(
        dna.children[1].spec,
        floatv(min_value=0.0, max_value=1.0)))
    self.assertTrue(symbolic.eq(
        dna.children[2].spec,
        custom()))

    dna = DNA((2, 0.0)).use_spec(spec)
    self.assertIs(dna.spec, spec)
    self.assertTrue(symbolic.eq(
        dna.children[0].spec,
        floatv(min_value=0.0, max_value=1.0, hints=2)))

    with self.assertRaisesRegex(
        ValueError, 'Argument \'spec\' must be a `pg.DNASpec` object.'):
      DNA(None).use_spec(1)
    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch'):
      DNA(None).use_spec(spec)
    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch'):
      DNA(1.5).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError,
        'Value of DNA is out of range according to the DNA spec'):
      DNA(3).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError, 'There is no DNA spec for child DNA values'):
      DNA((0, 0)).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError,
        'Number of elements in child templates .* does not match '
        'with the length of children .* from DNA'):
      DNA((1, 0)).use_spec(spec)

  def test_use_spec_on_manyof(self):
    spec = manyof(2, [
        constant(),
        constant(),
        constant(),
    ], sorted=True, distinct=True, hints=0)
    dna = DNA([0, 1]).use_spec(spec)
    self.assertIs(dna.spec, spec)
    child_spec = spec.clone(
        override=dict(num_choices=1, location='[0]', subchoice_index=0))
    self.assertTrue(symbolic.eq(dna.children[0].spec, child_spec))
    with self.assertRaisesRegex(
        ValueError, 'Cannot apply multi-choice DNA spec on value'):
      DNA(1).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError,
        'Number of choices .* does not match with the number '
        'of child values'):
      DNA([0, 0, 1]).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError, 'Child values .* are not sorted'):
      DNA([1, 0]).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError, 'Child values .* are not distinct'):
      DNA([0, 0]).use_spec(spec)

  def test_use_spec_on_float(self):
    spec = floatv(min_value=0.0, max_value=1.0, hints=1)
    self.assertIs(DNA(0.5).use_spec(spec).spec, spec)
    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch'):
      DNA(None).use_spec(spec)
    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch'):
      DNA(0).use_spec(spec)
    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch'):
      DNA([0, 1]).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError, 'DNA value should be no less than .*'):
      DNA(-0.1).use_spec(spec)
    with self.assertRaisesRegex(
        ValueError, 'DNA value should be no greater than .*'):
      DNA(2.).use_spec(spec)

  def test_use_spec_on_custom_decision_point(self):
    spec = custom()
    self.assertIs(DNA('abc').use_spec(spec).spec, spec)

    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch'):
      DNA(1).use_spec(Space([custom()]))

  def test_use_spec_on_complex(self):
    spec = Space([
        oneof([
            manyof(2, [
                constant(),
                floatv(0.0, 1.0),
                constant(),
                custom()
            ], sorted=True, distinct=True, location='b'),
            floatv(0.0, 1.0, location='c'),
            constant(),
        ], location='a'),
        floatv(-1.0, 0.0, location='d')
    ])
    dna = DNA([(0, [0, (1, 0.5, 'abc')]), -0.5]).use_spec(spec)
    self.assertIsNotNone(dna.spec)

    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch.'):
      DNA(1).use_spec(spec)

    with self.assertRaisesRegex(ValueError, 'DNA value type mismatch.'):
      DNA([0.5, -0.5]).use_spec(spec)

    with self.assertRaisesRegex(
        ValueError,
        'Number of choices .* does not match with the number of '
        'child values'):
      DNA([(0, 0), -0.5]).use_spec(spec)

    with self.assertRaisesRegex(
        ValueError,
        'Number of choices .* does not match with the number of '
        'child values'):
      DNA([(0, 2, 1), -0.5]).use_spec(spec)

    with self.assertRaisesRegex(
        ValueError,
        'Value of DNA is out of range according to the DNA spec.'):
      DNA([4, -0.5]).use_spec(spec)

    with self.assertRaisesRegex(
        ValueError, 'Encountered more than 1 value.'):
      DNA([1, [0.0, 1.0]]).use_spec(spec)

    with self.assertRaisesRegex(
        ValueError, 'Child values .* are not sorted'):
      DNA([(0, [(1, 0.5), 0]), -0.5]).use_spec(spec)

    with self.assertRaisesRegex(
        ValueError, 'Child values .* are not distinct'):
      DNA([(0, [(1, 0.2), (1, 0.5)]), -0.5]).use_spec(spec)

    with self.assertRaisesRegex(
        ValueError, 'Length of DNA child values .* is different from '
        'the number of elements.'):
      DNA([0, 1]).use_spec(Space([]))


class DNADictConversionTest(unittest.TestCase):
  """Tests for `pg.DNA.to_dict` and `pg.DNA.from_dict`."""

  def _dna(self):
    spec = Space([
        oneof([
            # NOTE(daiyip): multiple choices under conditions.
            manyof(3, [
                constant(),
                floatv(0.0, 1.0, name='c'),
                constant(),
                custom(name='f')
            ], literal_values=[
                123, 'Float(...)', 'xyz', 'Custom(...)'
            ], location='b', name='b'),
            constant(),
        ], literal_values=[
            'Something complex',
            '\'foo\''
        ], location='a', name='a'),
        floatv(-1.0, 0.0, location='d', name='d'),
        # NOTE(daiyip): multiple choices NOT under conditions.
        manyof(2, [
            constant(), constant(), constant()
        ], literal_values=[
            '\'a\'',
            0.333,
            '\'c\'',
        ], location='e', name='e')
    ])
    return DNA([(0, [0, (1, 0.5), (3, 'abc')]), -0.5, [0, 1]]).use_spec(spec)

  def test_id_and_value(self):
    self.assertEqual(
        self._dna().to_dict(),
        {
            'a': 0,
            'a[=0/2].b[0]': 0,
            'a[=0/2].b[1]': 1,
            'a[=0/2].b[1][=1/4]': 0.5,
            'a[=0/2].b[2]': 3,
            'a[=0/2].b[2][=3/4]': 'abc',
            'd': -0.5,
            'e[0]': 0,
            'e[1]': 1,
        })

  def test_id_and_dna(self):
    self.assertEqual(
        self._dna().to_dict(value_type='dna'),
        {
            'a': DNA((0, [0, (1, 0.5), (3, 'abc')])),
            'a[=0/2].b[0]': DNA(0),
            'a[=0/2].b[1]': DNA((1, 0.5)),
            'a[=0/2].b[1][=1/4]': DNA(0.5),
            'a[=0/2].b[2]': DNA((3, 'abc')),
            'a[=0/2].b[2][=3/4]': DNA('abc'),
            'd': DNA(-0.5),
            'e[0]': DNA(0),
            'e[1]': DNA(1),
        })

  def test_id_and_choice(self):
    self.assertEqual(
        self._dna().to_dict(value_type='choice'),
        {
            'a': '0/2',
            'a[=0/2].b[0]': '0/4',
            'a[=0/2].b[1]': '1/4',
            'a[=0/2].b[1][=1/4]': 0.5,
            'a[=0/2].b[2]': '3/4',
            'a[=0/2].b[2][=3/4]': 'abc',
            'd': -0.5,
            'e[0]': '0/3',
            'e[1]': '1/3',
        })

  def test_id_and_literal(self):
    self.assertEqual(
        self._dna().to_dict(value_type='literal'),
        {
            'a': 'Something complex',
            'a[=0/2].b[0]': 123,
            'a[=0/2].b[1]': 'Float(...)',
            'a[=0/2].b[1][=1/4]': 0.5,
            'a[=0/2].b[2]': 'Custom(...)',
            'a[=0/2].b[2][=3/4]': 'abc',
            'd': -0.5,
            'e[0]': '\'a\'',
            'e[1]': 0.333,
        })

  def test_id_and_choice_literal(self):
    self.assertEqual(
        self._dna().to_dict(value_type='choice_and_literal'),
        {
            'a': '0/2 (Something complex)',
            'a[=0/2].b[0]': '0/4 (123)',
            'a[=0/2].b[1]': '1/4 (Float(...))',
            'a[=0/2].b[1][=1/4]': 0.5,
            'a[=0/2].b[2]': '3/4 (Custom(...))',
            'a[=0/2].b[2][=3/4]': 'abc',
            'd': -0.5,
            'e[0]': '0/3 (\'a\')',
            'e[1]': '1/3 (0.333)',
        })

  def test_name_or_id_and_value(self):
    self.assertEqual(
        self._dna().to_dict(key_type='name_or_id'),
        {
            'a': 0,
            'b': [0, 1, 3],
            'c': 0.5,
            'd': -0.5,
            'e': [0, 1],
            'f': 'abc'
        })

  def test_name_or_id_and_dna(self):
    self.assertEqual(
        self._dna().to_dict(key_type='name_or_id', value_type='dna'),
        {
            'a': DNA((0, [0, (1, 0.5), (3, 'abc')])),
            'b': [DNA(0), DNA((1, 0.5)), DNA((3, 'abc'))],
            'c': DNA(0.5),
            'd': DNA(-0.5),
            'e': [DNA(0), DNA(1)],
            'f': DNA('abc'),
        })

  def test_name_or_id_and_choice(self):
    self.assertEqual(
        self._dna().to_dict(key_type='name_or_id', value_type='choice'),
        {
            'a': '0/2',
            'b': ['0/4', '1/4', '3/4'],
            'c': 0.5,
            'd': -0.5,
            'e': ['0/3', '1/3'],
            'f': 'abc'
        })

  def test_name_or_id_and_literal(self):
    self.assertEqual(
        self._dna().to_dict(key_type='name_or_id', value_type='literal'),
        {
            'a': 'Something complex',
            'b': [123, 'Float(...)', 'Custom(...)'],
            'c': 0.5,
            'd': -0.5,
            'e': ['\'a\'', 0.333],
            'f': 'abc'
        })

  def test_name_or_id_and_choice_literal(self):
    self.assertEqual(
        self._dna().to_dict(
            key_type='name_or_id',
            value_type='choice_and_literal'),
        {
            'a': '0/2 (Something complex)',
            'b': ['0/4 (123)', '1/4 (Float(...))', '3/4 (Custom(...))'],
            'c': 0.5,
            'd': -0.5,
            'e': ['0/3 (\'a\')', '1/3 (0.333)'],
            'f': 'abc'
        })

  def test_dna_spec_and_value(self):
    dna = self._dna()
    named = dna.spec.named_decision_points
    self.assertEqual(
        dna.to_dict(key_type='dna_spec'),
        {
            named['a']: 0,
            named['b'][0]: 0,
            named['b'][1]: 1,
            named['b'][2]: 3,
            named['c'][1]: 0.5,
            named['d']: -0.5,
            named['e'][0]: 0,
            named['e'][1]: 1,
            named['f'][2]: 'abc',
        })

  def test_dna_spec_and_dna(self):
    dna = self._dna()
    named = dna.spec.named_decision_points
    self.assertEqual(
        dna.to_dict(key_type='dna_spec', value_type='dna'),
        {
            named['a']: DNA((0, [0, (1, 0.5), (3, 'abc')])),
            named['b'][0]: DNA(0),
            named['b'][1]: DNA((1, 0.5)),
            named['b'][2]: DNA((3, 'abc')),
            named['c'][1]: DNA(0.5),
            named['d']: DNA(-0.5),
            named['e'][0]: DNA(0),
            named['e'][1]: DNA(1),
            named['f'][2]: DNA('abc')
        })

  def test_dna_spec_and_choice(self):
    dna = self._dna()
    named = dna.spec.named_decision_points
    self.assertEqual(
        dna.to_dict(key_type='dna_spec', value_type='choice'),
        {
            named['a']: '0/2',
            named['b'][0]: '0/4',
            named['b'][1]: '1/4',
            named['b'][2]: '3/4',
            named['c'][1]: 0.5,
            named['d']: -0.5,
            named['e'][0]: '0/3',
            named['e'][1]: '1/3',
            named['f'][2]: 'abc'
        })

  def test_dna_spec_and_literal(self):
    dna = self._dna()
    named = dna.spec.named_decision_points
    self.assertEqual(
        dna.to_dict(key_type='dna_spec', value_type='literal'),
        {
            named['a']: 'Something complex',
            named['b'][0]: 123,
            named['b'][1]: 'Float(...)',
            named['b'][2]: 'Custom(...)',
            named['c'][1]: 0.5,
            named['d']: -0.5,
            named['e'][0]: '\'a\'',
            named['e'][1]: 0.333,
            named['f'][2]: 'abc'
        })

  def test_dna_spec_and_choice_literal(self):
    dna = self._dna()
    named = dna.spec.named_decision_points
    self.assertEqual(
        dna.to_dict(key_type='dna_spec', value_type='choice_and_literal'),
        {
            named['a']: '0/2 (Something complex)',
            named['b'][0]: '0/4 (123)',
            named['b'][1]: '1/4 (Float(...))',
            named['b'][2]: '3/4 (Custom(...))',
            named['c'][1]: 0.5,
            named['d']: -0.5,
            named['e'][0]: '0/3 (\'a\')',
            named['e'][1]: '1/3 (0.333)',
            named['f'][2]: 'abc'
        })

  def test_collapse_multi_choices(self):
    self.assertEqual(
        self._dna().to_dict(multi_choice_key='parent'),
        {
            'a': 0,
            'a[=0/2].b': [0, 1, 3],
            'a[=0/2].b[1][=1/4]': 0.5,
            'a[=0/2].b[2][=3/4]': 'abc',
            'd': -0.5,
            'e': [0, 1]
        })

  def test_keep_both_parent_and_subchoices(self):
    dna = self._dna()
    named = dna.spec.named_decision_points
    self.assertEqual(
        dna.to_dict(key_type='dna_spec', value_type='dna',
                    multi_choice_key='both'),
        {
            named['a']: DNA((0, [0, (1, 0.5), (3, 'abc')])),
            named['b'][0].parent_spec: [DNA(0),
                                        DNA((1, 0.5)),
                                        DNA((3, 'abc'))],
            named['b'][0]: DNA(0),
            named['b'][1]: DNA((1, 0.5)),
            named['b'][2]: DNA((3, 'abc')),
            named['c'][1]: DNA(0.5),
            named['d']: DNA(-0.5),
            named['e'][0].parent_spec: [DNA(0), DNA(1)],
            named['e'][0]: DNA(0),
            named['e'][1]: DNA(1),
            named['f'][2]: DNA('abc'),
        })

  def test_include_inactive_decisions(self):
    self.assertEqual(
        self._dna().to_dict(include_inactive_decisions=True),
        {
            'a': 0,
            'a[=0/2].b[0]': 0,
            'a[=0/2].b[0][=1/4]': None,
            'a[=0/2].b[0][=3/4]': None,
            'a[=0/2].b[1]': 1,
            'a[=0/2].b[1][=1/4]': 0.5,
            'a[=0/2].b[1][=3/4]': None,
            'a[=0/2].b[2]': 3,
            'a[=0/2].b[2][=1/4]': None,
            'a[=0/2].b[2][=3/4]': 'abc',
            'd': -0.5,
            'e[0]': 0,
            'e[1]': 1
        })

  def test_include_inactive_decisions_and_collapse_multi_choices(self):
    self.assertEqual(
        self._dna().to_dict(
            value_type='dna',
            multi_choice_key='parent',
            include_inactive_decisions=True),
        {
            'a': DNA((0, [0, (1, 0.5), (3, 'abc')])),
            'a[=0/2].b': [DNA(0),
                          DNA((1, 0.5)),
                          DNA((3, 'abc'))],
            'a[=0/2].b[0][=1/4]': None,
            'a[=0/2].b[0][=3/4]': None,
            'a[=0/2].b[1][=1/4]': DNA(0.5),
            'a[=0/2].b[1][=3/4]': None,
            'a[=0/2].b[2][=1/4]': None,
            'a[=0/2].b[2][=3/4]': DNA('abc'),
            'd': DNA(-0.5),
            'e': [DNA(0), DNA(1)],
        })

  def test_from_dict(self):
    dna = self._dna()
    spec = dna.spec
    for kt in ['id', 'name_or_id', 'dna_spec']:
      for vt in ['value', 'dna', 'choice', 'literal', 'choice_and_literal']:
        for mk in ['subchoice', 'parent', 'both']:
          for inactive in [True, False]:
            self.assertEqual(
                DNA.from_dict(
                    dna.to_dict(key_type=kt, value_type=vt,
                                multi_choice_key=mk,
                                include_inactive_decisions=inactive),
                    spec, use_ints_as_literals=(vt == 'literal')),
                dna)

  def test_from_dict_with_filter_fn(self):
    self.assertEqual(
        self._dna().to_dict(
            value_type='dna',
            multi_choice_key='parent',
            include_inactive_decisions=True,
            filter_fn=lambda x: x.is_custom_decision_point),
        {
            'a[=0/2].b[0][=3/4]': None,
            'a[=0/2].b[1][=3/4]': None,
            'a[=0/2].b[2][=3/4]': DNA('abc'),
        })

  def test_bad_to_dict(self):
    dna = self._dna()
    with self.assertRaisesRegex(
        ValueError, '\'key_type\' must be either .*'):
      dna.to_dict(key_type='foo')

    with self.assertRaisesRegex(
        ValueError, '\'value_type\' must be either .*'):
      dna.to_dict(value_type='foo')

    with self.assertRaisesRegex(
        ValueError, '\'multi_choice_key\' must be either .*'):
      dna.to_dict(multi_choice_key='foo')

    with self.assertRaisesRegex(
        ValueError, '.* is not bound with a DNASpec'):
      DNA(0).to_dict()

  def test_bad_from_dict(self):
    # Categorical.
    spec = Space([oneof([constant(), constant()], location='a')])
    with self.assertRaisesRegex(ValueError, 'Candidate index out of range'):
      DNA.from_dict({'a': '2/2'}, spec)

    # Numerical.
    spec = Space([floatv(location='a', min_value=0.0, max_value=1.0)])
    with self.assertRaisesRegex(
        ValueError, 'Value for .* is not found in the dictionary .*'):
      DNA.from_dict({'b': '0/3'}, spec)
    with self.assertRaisesRegex(
        ValueError, 'The decision for .* should be no less than .*'):
      DNA.from_dict({'a': -1.0}, spec)
    with self.assertRaisesRegex(
        ValueError, 'The decision for .* should be no greater than .*'):
      DNA.from_dict({'a': 2.0}, spec)

    # Custom decision point.
    custom_spec = Space([custom(location='a')])
    with self.assertRaisesRegex(
        ValueError, 'The decision for .* should be a string'):
      DNA.from_dict({'a': 1}, custom_spec)

    # Complex case:
    spec = self._dna().spec
    with self.assertRaisesRegex(
        ValueError, 'Value for .* is not found in the dictionary .*'):
      DNA.from_parameters({'x': 1}, spec)
    with self.assertRaisesRegex(
        ValueError, 'There is no candidate in .*'):
      DNA.from_parameters({'a': 'foo'}, spec)
    with self.assertRaisesRegex(
        ValueError,
        'Number of candidates .* for Choice .* does not match with DNASpec'):
      DNA.from_parameters({'a': '0/3'}, spec)

  def test_from_parameters(self):
    dna = self._dna()
    spec = dna.spec

    dna2 = DNA.from_parameters(dna.parameters(), spec)
    self.assertEqual(dna, dna2)

    dna3 = DNA.from_parameters(
        dna.parameters(use_literal_values=True), spec, use_literal_values=True)
    self.assertEqual(dna, dna3)


class ConditionalKeyTest(unittest.TestCase):
  """Tests for ConditionalKey."""

  def test_basics(self):
    key = ConditionalKey(1, 5)
    self.assertEqual(key.index, 1)
    self.assertEqual(key.num_choices, 5)

  def test_to_str(self):
    key = utils.KeyPath(['a', ConditionalKey(1, 5), 'b'])
    self.assertEqual(str(key), 'a[=1/5].b')


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base for DNA generators."""

from typing import Callable, Iterable, Iterator, Optional, Tuple, Union

from pyglove.core import symbolic
from pyglove.core.geno.base import DNA
from pyglove.core.geno.base import DNASpec


class DNAGenerator(symbolic.Object):
  """Base class for DNA generator.

  A DNA generator is an object that produces :class:`pyglove.DNA`, and
  optionally takes feedback from the caller to improve its future proposals.

  To implement a DNA generator, the user must implement the `_propose` method,
  and can optionally override the `_setup`, `_feedback` and `_replay` methods.

   * Making proposals (Required): This method defines what to return as the
     next DNA from the generator, users MUST override the `_propose` method to
     implement this logic. `_propose` can raise `StopIteration` when no more
     DNA can be produced.

   * Custom setup (Optional): Usually a DNAGenerator subclass has its internal
     state, which can be initialized when the search space definition is
     attached to the DNAGenerator. To do so, the user can override the `_setup`
     method, in which we can access the search space definition (DNASpec object)
     via `self.dna_spec`.

   * Taking feedback (Optional): A DNAGenerator may take feedbacks from the
     caller on the fitness of proposed DNA to improve future proposals. The
     fitness is measured by a reward (a float number as the measure of a single
     objective, or a tuple of float numbers as the measure for multiple
     objectives). The user should override the `_feedback` method to implement
     such logics. If the reward is for multiple objectives. The user should
     override the `multi_objective` property to return True.

   * State recovery (Optional): DNAGenerator was designed with distributed
     computing in mind, in which a process can be preempted or killed
     unexpectedly. Therefore, a DNAGenerator should be able to recover its
     state from historical proposals and rewards. The `recover` method was
     introduced for such purpose, whose default implementation is to replay the
     history through the `_feedback` method. If the user has a custom replay
     logic other than `_feedback`, they should override the `_replay` method.
     In some use cases, the user may want to implement their own checkpointing
     logic. In such cases, the user can override the `recover` method as a
     no-op. As aside note, the `recover` method will be called by the tuning
     backend (see `tuning.py`) after `setup` but before `propose`.

  See also:

   * :class:`pyglove.geno.Sweeping`
   * :class:`pyglove.geno.Random`
   * :func:`pyglove.geno.dna_generator`
  """

  def setup(self, dna_spec: DNASpec) -> None:
    """Setup DNA spec."""
    self._dna_spec = dna_spec
    self._num_proposals = 0
    self._num_feedbacks = 0
    self._setup()

  def _setup(self) -> None:
    """Subclass should override this for adding additional setup logics."""

  @property
  def multi_objective(self) -> bool:
    """If True, current DNA generator supports multi-objective optimization."""
    return False

  @property
  def needs_feedback(self) -> bool:
    """Returns True if the DNAGenerator needs feedback."""
    return self._feedback.__code__ is not DNAGenerator._feedback.__code__  # pytype: disable=attribute-error

  @property
  def dna_spec(self) -> Optional[DNASpec]:
    return getattr(self, '_dna_spec', None)

  @property
  def num_proposals(self):
    """Get number of proposals that are already produced."""
    return self._num_proposals

  @property
  def num_feedbacks(self):
    """Get number of proposals whose feedback are provided."""
    return self._num_feedbacks

  def propose(self) -> DNA:
    """Propose a DNA to evaluate."""
    dna = self._propose()
    self._num_proposals += 1
    return dna

  def _propose(self) -> DNA:
    """Actual propose method which should be implemented by the child class."""
    raise NotImplementedError()

  def feedback(self, dna: DNA, reward: Union[float, Tuple[float]]) -> None:
    """Feedback a completed trial to the algorithm.

    Args:
      dna: a DNA object.
      reward: reward for the DNA. It is a float if `self.multi_objective`
        returns False, otherwise it's a tuple of floats.
    """
    if self.needs_feedback:
      if self.multi_objective and isinstance(reward, float):
        reward = (reward,)
      elif not self.multi_objective and isinstance(reward, tuple):
        if len(reward) != 1:
          raise ValueError(
              f'{self!r} is single objective, but the reward {reward!r} '
              f'contains multiple objectives.')
        reward = reward[0]
      self._feedback(dna, reward)
    self._num_feedbacks += 1

  def _feedback(self, dna: DNA, reward: Union[float, Tuple[float]]) -> None:
    """Actual feedback method which should be implemented by the child class.

    The default implementation is no-op.

    Args:
      dna: a DNA object.
      reward: reward for the DNA. It is a float if `self.multi_objective`
        returns False, otherwise it's a tuple of floats.
    """

  def recover(
      self,
      history: Iterable[Tuple[DNA, Union[None, float, Tuple[float]]]]
      ) -> None:
    """Recover states by replaying the proposal history.

    NOTE: `recover` will always be called before first `propose` and could be
    called multiple times if there are multiple source of history, e.g: trials
    from a previous study and existing trials from current study.

    Args:
      history: An iterable object that consists of historically proposed DNA
        with its reward. The reward will be None if it is not yet provided
        (via feedback).
    """
    for i, (dna, reward) in enumerate(history):
      self._replay(i, dna, reward)
      self._num_proposals += 1
      if reward is not None:
        self._num_feedbacks += 1

  def _replay(
      self,
      trial_id: int,
      dna: DNA,
      reward: Union[None, float, Tuple[float]]):
    """Replay a single DNA from the history for state recovery.

    The default implementation to call `DNAGenerator._feedback`. Subclasses that
    have states and can be recovered from replaying the history should override
    this method. See class `Sweeping` as an example.

    Args:
      trial_id: A zero-based integer as the trial ID for the DNA.
      dna: A historically proposed DNA.
      reward: The reward for the DNA. If None, the reward is not yet fed back
        to the optimizer.
    """
    del trial_id
    if reward is not None:
      self._feedback(dna, reward)

  def __iter__(self) -> Iterator[
      Union[DNA,
            Tuple[DNA, Callable[[Union[float, Tuple[float]]], None]]]]:
    """Iterates DNA generated from current DNAGenerator.

    NOTE(daiyip): `setup` needs to be called first before a DNAGenerator can
    be iterated.

    Yields:
      A tuple of (DNA, feedback) if current DNAGenerator requires feedback,
        otherwise DNA.
    """
    while True:
      try:
        dna = self.propose()
        if self.needs_feedback:
          feedback = lambda r: self.feedback(dna, r)
          yield (dna, feedback)
        else:
          yield dna
      except StopIteration:
        break


def dna_generator(func: Callable[[DNASpec], Iterator[DNA]]):
  """Decorator that converts a generation function to a DNAGenerator class.

  Example::

    # A DNA generator that reads DNA from file.

    def from_file(filepath):
      @pg.geno.dna_generator
      def file_based_dna_generator(dna_spec):
        dna_list = pg.load(filepath)
        for dna in dna_list:
          dna.use_spec(dna_spec)
          yield dna

      return file_based_dna_generator

  See also: :class:`pyglove.DNAGenerator`

  Args:
    func: the generation function in signature:
      `(DNASpec) -> Iterator[DNA]`

  Returns:
    A DNAGenerator class.
  """

  class SimpleDNAGenerator(DNAGenerator):
    """Simple DNA generator."""

    def _setup(self):
      self._iterator = func(self.dna_spec)
      self._error = None

    def _propose(self) -> DNA:
      if self._error is not None:
        raise ValueError(
            f'Error happened earlier: {self._error}') from self._error
      try:
        return next(self._iterator)
      except Exception as e:
        if not isinstance(e, StopIteration):
          self._error = e
        raise

  return SimpleDNAGenerator



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Genotype for categorical decisions."""

import random
import re
import types
from typing import Any, List, Optional, Union

from pyglove.core import symbolic
from pyglove.core import typing as pg_typing
from pyglove.core import utils
from pyglove.core.geno.base import DecisionPoint
from pyglove.core.geno.base import DNA
from pyglove.core.geno.base import DNASpec
from pyglove.core.geno.space import Space


_CHOICE_REGEX = re.compile(r'^(\d+)/(\d+)$')
_CHOICE_AND_LITERAL_REGEX = re.compile(r'^(\d+)/(\d+) \((.*)\)$', re.S)


@symbolic.members(
    [
        ('num_choices', pg_typing.Int(min_value=1, default=1),
         'Number of choices to make.'),
        ('candidates', pg_typing.List(pg_typing.Object(Space), min_size=1),
         ('A list of Spaces for every candidate, which may contain '
          'child nodes to form sub (conditional) search spaces.')),
        ('literal_values', pg_typing.List(
            pg_typing.Union([
                pg_typing.Str(), pg_typing.Int(), pg_typing.Float()]),
            min_size=1).noneable(),
         'Optional literal value for each candidate. Used for debugging.'),
        ('distinct', pg_typing.Bool(True), 'Whether choices are distinct.'),
        ('sorted', pg_typing.Bool(False),
         ('Whether choices are sorted. The order key is the index of the '
          'candidate in the `candidates` field instead of the candidate\'s '
          'value.')),
        ('subchoice_index', pg_typing.Int(min_value=0).noneable(),
         ('Index of current choice as a subchoice of a parent multi-choice. '
          'If None, current choice is either a single choice or a root spec '
          'for a multi-choice.'))
    ],
    init_arg_list=[
        'num_choices', 'candidates', 'literal_values',
        'distinct', 'sorted', 'subchoice_index',
        'hints', 'location', 'name'],
    # TODO(daiyip): For backward compatibility.
    # Move this to additional keys later.
    serialization_key='pyglove.generators.geno.Choices',
    additional_keys=['geno.Choices']
)
class Choices(DecisionPoint):
  """Represents a single or multiple choices from a list of candidates.

  Example::

    # Create a single choice with a nested subspace for candidate 2 (0-based).
    pg.geno.oneof([
        pg.geno.constant(),
        pg.geno.constant(),
        pg.geno.space([
            pg.geno.floatv(0.1, 1.0)
            pg.geno.manyof(2, [
                pg.geno.constant(),
                pg.geno.constant(),
                pg.geno.constant()
            ])
        ])
    ])

  See also: :func:`pyglove.geno.oneof`, :func:`pyglove.geno.manyof`.
  """

  def _on_bound(self):
    super()._on_bound()
    if self.distinct and self.num_choices > len(self.candidates):
      raise ValueError(
          f'There are not enough candidates ({len(self.candidates)}) to make '
          f'{self.num_choices} distinct choices.')
    if self.num_choices > 1 and self.subchoice_index is not None:
      raise ValueError(
          f'Multi-choice spec cannot be a subchoice. '
          f'Encountered: {self!r}.')
    if (self.literal_values is not None and
        len(self.literal_values) != len(self.candidates)):
      raise ValueError(
          f'The length of \'candidates\' ({len(self.candidates)}) '
          f'should be equal to the length of \'literal_values\' '
          f'({len(self.literal_values)}).')

    if self.literal_values is not None:
      # NOTE(daiyip): old PyGlove client will generate literal value in
      # format of 'choice_and_literal'. For backward compatibility, we will
      # convert the old format to the new format. We can deprecate this logic
      # in future.
      literal_values = []
      updated = False
      for i, literal in enumerate(self.literal_values):
        if isinstance(literal, str):
          r = _CHOICE_AND_LITERAL_REGEX.match(literal)
          if (r
              and int(r.group(1)) == i
              and int(r.group(2)) == len(self.candidates)):
            literal = r.group(3)
            updated = True
        literal_values.append(literal)
      if updated:
        self.rebind(literal_values=literal_values, skip_notification=True)

      # Store the literal to index mapping for quick candidate index lookup.
      self._literal_index = {v: i for i, v in enumerate(self.literal_values)}
    else:
      self._literal_index = {}

    self._space_size = None

    # Automatically set the candidate index for template.
    for i, c in enumerate(self.candidates):
      c.rebind(index=i, skip_notification=True, raise_on_no_change=False)

    # Create sub choice specs and index decision points.
    if self.num_choices > 1 and not self.is_subchoice:
      subchoice_specs = []
      self._decision_points = []
      for i in range(self.num_choices):
        subchoice_spec = Choices(
            subchoice_index=i,
            location=utils.KeyPath(i),
            num_choices=1,
            candidates=self.candidates,
            literal_values=self.literal_values,
            distinct=self.distinct,
            sorted=self.sorted,
            name=self.name,
            hints=self.hints,
        )
        self._decision_points.extend(subchoice_spec.decision_points)
        subchoice_specs.append(subchoice_spec)
      self._subchoice_specs = symbolic.List(subchoice_specs)
      self._subchoice_specs.sym_setparent(self)
    else:
      self._subchoice_specs = None
      self._decision_points = [self]
      for c in self.candidates:
        self._decision_points.extend(c.decision_points)

  def _update_children_paths(
      self, old_path: utils.KeyPath, new_path: utils.KeyPath
  ):
    """Trigger path change for subchoices so their IDs can be invalidated."""
    super()._update_children_paths(old_path, new_path)
    if self._subchoice_specs:
      for i, spec in enumerate(self._subchoice_specs):
        spec.sym_setpath(new_path + i)

  def subchoice(self, index: int) -> 'Choices':
    """Returns spec for choice i."""
    if self.is_subchoice:
      raise ValueError(
          f'\'subchoice\' should not be called on a subchoice of a '
          f'multi-choice. Encountered: {self!r}.')
    if self.num_choices == 1:
      return self
    return self._subchoice_specs[index]

  @property
  def choice_specs(self) -> List['Choices']:
    """Returns all choice specs."""
    return [self.subchoice(i) for i in range(self.num_choices)]

  @property
  def is_categorical(self) -> bool:
    """Returns True if current node is a categorical choice."""
    return True

  @property
  def is_subchoice(self) -> bool:
    """Returns if current choice is a subchoice of a multi-choice."""
    return self.subchoice_index is not None

  @property
  def is_numerical(self) -> bool:
    """Returns True if current node is numerical decision."""
    return False

  @property
  def is_custom_decision_point(self) -> bool:
    """Returns True if current node is a custom decision point."""
    return False

  def format_candidate(
      self,
      index: int,
      display_format: str = 'choice_and_literal') -> Union[str, int, float]:
    """Get a formatted candidate value by index.

    Args:
      index: The index of the candidate to format.
      display_format: One of 'choice', 'literal' and 'choice_and_literal' as
        the output format for human consumption.

    Returns:
      A int, float or string that represent the candidate based on the
        display format.
    """
    if display_format not in ['choice', 'literal', 'choice_and_literal']:
      raise ValueError(
          f'`display_format` must be either \'choice\', \'literal\', or '
          f'\'choice_and_literal\'. Encountered: {display_format!r}.')
    if self.literal_values:
      if display_format == 'literal':
        return self.literal_values[index]
      elif display_format == 'choice_and_literal':
        return f'{index}/{len(self.candidates)} ({self.literal_values[index]})'
    return f'{index}/{len(self.candidates)}'

  def candidate_index(self, choice_value: Union[str, int, float]) -> int:
    """Returns the candidate index of a choice value.

    Args:
      choice_value: Choice value can be:
        a (integer, float or string) as a candidate's literal value,
        or a text in the format "<index>/<num_candidates>"
        or a text in the format "<index>/<num_candidates> (<literal>)".

    Returns:
      The index of chosen candidate.

    Raises:
      ValueError: `choice_value` is not a valid index or it does not matches
      any candidate's literal value.
    """
    def _index_from_literal(value: Union[str, int, float]) -> int:
      index = self._literal_index.get(value)
      if index is None:
        raise ValueError(
            f'There is no candidate in {self!r} with literal value {value!r}.')
      return index

    index = None
    literal = None
    num_candidates = len(self.candidates)
    if isinstance(choice_value, str):
      r = _CHOICE_AND_LITERAL_REGEX.match(choice_value)
      if r:
        index = int(r.group(1))
        num_candidates = int(r.group(2))
        literal = r.group(3)
      else:
        r = _CHOICE_REGEX.match(choice_value)
        if r:
          index = int(r.group(1))
          num_candidates = int(r.group(2))
        else:
          index = _index_from_literal(choice_value)
    elif isinstance(choice_value, (int, float)):
      index = _index_from_literal(choice_value)
    else:
      raise ValueError(
          f'The value for Choice \'{self.id}\' should be either an integer, '
          f'a float or a string. Encountered: {choice_value!r}.')
    if index < -len(self.candidates) or index >= len(self.candidates):
      raise ValueError(
          f'Candidate index out of range at choice \'{self.name or self.id}\'. '
          f'Index={index}, Number of candidates={len(self.candidates)}.')
    if num_candidates != len(self.candidates):
      raise ValueError(
          f'Number of candidates ({num_candidates}) for Choice '
          f'\'{self.name or self.id}\' does not match with '
          f'DNASpec ({len(self.candidates)}).')
    if (literal is not None
        and (not self.literal_values
             or literal != str(self.literal_values[index]))):
      raise ValueError(
          f'The literal value from the input ({literal!r}) does not match '
          f'with the literal value ({self.literal_values[index]!r}) from the '
          f'chosen candidate (index={index}).')
    return index

  def validate(self, dna: DNA) -> None:
    """Validate whether a DNA value conforms to this spec."""
    if self.num_choices == 1:
      # For single choice, choice is decoded from dna.value.
      # children should be decoded by chosen candidate.
      if dna.value is None and dna.children:
        raise ValueError(
            f'Expect an integer for a single choice, but encountered a list '
            f'({dna.to_numbers(flatten=False)}). '
            f'Location: {self.location.path}.')
      if not isinstance(dna.value, int):
        raise ValueError(
            f'Expect an integer for a single choice, but encountered: '
            f'{dna.value!r}, Location: {self.location.path}.')
      if dna.value >= len(self.candidates):
        raise ValueError(
            f'Choice out of range. Value: {dna.value}, '
            f'Candidates: {len(self.candidates)}, '
            f'Location: {self.location.path}.')
      chosen = self.candidates[dna.value]
      if chosen.is_constant and dna.children:
        raise ValueError(
            f'Child DNA ({dna.children!r}) provided while there is no child '
            f'decision point.')
      elif not chosen.is_constant and not dna.children:
        raise ValueError(
            f'No child DNA provided for child space {chosen!r}.')
      chosen.validate(DNA(None, dna.children))
    else:
      # For multiple choices, choice is encoded in DNA.children.
      # dna.value could be None (Choices as the only encoder in root template)
      # or int (parent choice).
      if len(dna.children) != self.num_choices:
        raise ValueError(
            f'Number of DNA child values does not match the number of choices. '
            f'Child values: {dna.children!r}, Choices: {self.num_choices}, '
            f'Location: {self.location.path}.')
      if self.distinct or self.sorted:
        sub_dna_values = [s.value for s in dna]
        if self.distinct and len(set(sub_dna_values)) != len(dna.children):
          raise ValueError(
              f'DNA child values should be distinct. '
              f'Encountered: {sub_dna_values}, Location: {self.location.path}.')
        if self.sorted and sorted(sub_dna_values) != sub_dna_values:
          raise ValueError(
              f'DNA child values should be sorted. '
              f'Encountered: {sub_dna_values}, Location: {self.location.path}.')
      for i, sub_dna in enumerate(dna):
        sub_location = utils.KeyPath(i, self.location)
        if not isinstance(sub_dna.value, int):
          raise ValueError(
              f'Choice value should be int. Encountered: {sub_dna.value}, '
              f'Location: {sub_location.path}.')
        if sub_dna.value >= len(self.candidates):
          raise ValueError(
              f'Choice out of range. Value: {sub_dna.value}, '
              f'Candidates: {len(self.candidates)}, '
              f'Location: {sub_location.path}.')
        self.candidates[sub_dna.value].validate(DNA(None, sub_dna.children))

  @property
  def decision_points(self) -> List[DecisionPoint]:
    """Returns all decision points in their declaration order.

    Returns:
      All decision points in current space. For multi-choices, the sub-choice
      objects will be returned. Users can call `spec.parent_choice` to access
      the parent multi-choice node.
    """
    return self._decision_points

  @property
  def space_size(self) -> int:
    """Returns the search space size."""
    if self._space_size is None:
      sub_space_sizes = [c.space_size for c in self.candidates]
      if any([v == -1 for v in sub_space_sizes]):
        self._space_size = -1
      else:
        def _space_size(s: List[int], k: int) -> int:
          """Returns space size of k choices from a list of sub-space sizes."""
          if k == 0:
            return 1
          elif k == 1:
            return sum(s)
          elif k > len(s) and self.distinct:
            return 0
          elif len(s) == 1:
            assert not self.distinct
            return s[0] ** k  # pytype: disable=bad-return-type
          elif self.distinct and self.sorted:
            # When choice is distinct and sorted, current chosen item
            # must appear at the front.
            return s[0] * _space_size(s[1:], k - 1) + _space_size(s[1:], k)
          elif self.distinct:
            # When choice is distinct but not sorted, current chosen item
            # could appear at all `k` positions.
            return s[0] * k * _space_size(s[1:], k - 1) + _space_size(s[1:], k)
          elif self.sorted:
            # When choice is sorted but not distinct, the space size is the sum
            # of k + 1 terms, where the i'th term (0 <= i <= k) is the space
            # size of selecting s[0] for i times multiplying the space size of
            # choosing k - i items from s[1:].
            size = 0
            for i in range(k + 1):
              size += (s[0] ** i) * _space_size(s[1:], k - i)
            return size  # pytype: disable=bad-return-type  # always-use-return-annotations
          else:
            # When choice is neither distinct nor sorted,
            return _space_size(s, 1) ** k  # pytype: disable=bad-return-type  # always-use-return-annotations
        self._space_size = _space_size(sub_space_sizes, self.num_choices)
    return self._space_size

  def _next_dna(self, dna: Optional[DNA] = None) -> Optional[DNA]:
    """Returns the next DNA."""
    if dna is None:
      choice_dna_list = []
      for i in range(self.num_choices):
        choice = i if self.distinct else 0
        dna = self.candidates[choice].first_dna(attach_spec=False)
        choice_dna_list.append(DNA(choice, [dna]))
      return DNA(None, choice_dna_list)

    if self.num_choices == 1:
      # NOTE(daiyip): the DNA represenation of Choices should be
      # DNA(value=None, children=[dna_for_choice1, dna_for_choice2, ...]
      # However, `DNA(value=None, children=[choice1])` will be collapsed
      # into `choice1` for a more compact representation when `num_choices`
      # is equal to 1.
      parent_choice_value = None
      choice_dna_list = [dna]
    else:
      if len(dna.children) != self.num_choices:
        raise ValueError(
            f'Expect {self.num_choices} choices but encountered '
            f'{len(dna.children)} sub-DNA ({dna.children!r}).')
      parent_choice_value = dna.value
      choice_dna_list = dna.children

    def next_value_for_choice(
        prior_choices: List[int], current_choice: int) -> Optional[int]:
      """Get the next value conditioned by prior choices and current value."""
      n = len(self.candidates)
      next_value = current_choice + 1
      if self.distinct:
        possible_choices = set(range(next_value, n))
        possible_choices -= set(prior_choices)
        next_value = min(possible_choices) if possible_choices else n
      return next_value if next_value < n else None

    def min_remaining_choices(prior_choices: List[int]) -> Optional[List[int]]:
      """Get the minimal remaining choices conditioned by prior choices.."""
      if self.sorted and prior_choices:
        possible_choices = set(range(prior_choices[-1], len(self.candidates)))
      else:
        possible_choices = set(range(len(self.candidates)))
      if self.distinct:
        possible_choices -= set(prior_choices)

      remaining_choices = []
      for _ in range(self.num_choices - len(prior_choices)):
        if not possible_choices:
          return None
        next_choice = min(possible_choices)
        if self.distinct:
          possible_choices.remove(next_choice)
        remaining_choices.append(next_choice)
      return remaining_choices

    # Now we want to increment the DNA from right to left (last choice first),
    # which means we will increment the sub-DNA if possible, or we will have
    # to increment the choice value with its first DNA. If we can increment
    # neither the sub-DNA nor the choice value, we have reached the end of the
    # space represented by the DNASpec.
    for choice_id in reversed(range(self.num_choices)):
      choice_dna = choice_dna_list[choice_id]

      if (not isinstance(choice_dna.value, int)
          or choice_dna.value < 0 or choice_dna.value >= len(self.candidates)):
        raise ValueError(
            f'Choice value ({choice_dna.value}) is out of range '
            f'([0, {len(self.candidates)}]). DNASpec: {self!r}.')

      subspace_dna = DNA(None, choice_dna.children)
      subspace_next_dna = self.candidates[choice_dna.value].next_dna(
          subspace_dna, attach_spec=False)

      updated_current_choice = False
      prior_choices = [choice_dna_list[i].value for i in range(choice_id)]
      if subspace_next_dna is not None:
        # The sub-DNA is incremented, thus we can keep using the same choice
        # value.
        new_choice_dna = DNA(choice_dna.value, [subspace_next_dna])
        updated_current_choice = True
      else:
        # We need to find the next valid value for current choice.
        new_choice_value = next_value_for_choice(
            prior_choices, choice_dna.value)
        if new_choice_value is not None:
          new_choice_dna = DNA(
              new_choice_value,
              [self.candidates[new_choice_value].first_dna(
                  attach_spec=False)])
          updated_current_choice = True

      if updated_current_choice:
        # We found next sub-DNA at choice i, therefore we can reuse sub-DNA
        # from choice [0...i-1], and find remaining choices for choices
        # [i+1...k].
        remaining_choices = min_remaining_choices(
            prior_choices + [new_choice_dna.value])

        if remaining_choices is not None:
          subdna_list = choice_dna_list[:choice_id] + [new_choice_dna] + [
              DNA(v, [self.candidates[v].first_dna(attach_spec=False)])
              for v in remaining_choices
          ]
          return DNA(parent_choice_value, subdna_list)
    return None

  def _random_dna(
      self,
      random_generator: Union[random.Random, types.ModuleType],
      previous_dna: Optional[DNA]) -> DNA:
    """Returns a random DNA based on current spec."""
    if self.distinct:
      choices = random_generator.sample(
          list(range(len(self.candidates))), self.num_choices)
    else:
      choices = [random_generator.randint(0, len(self.candidates) - 1)
                 for _ in range(self.num_choices)]
    if self.sorted:
      choices = sorted(choices)

    # Figure out previous DNAs.
    if previous_dna is None:
      child_dnas = [None] * len(choices)
    else:
      if self.num_choices == 1:
        choices_dnas = [previous_dna]
      else:
        choices_dnas = list(previous_dna.children)
      assert len(choices_dnas) == self.num_choices, (self, choices_dnas)
      child_dnas = []
      for i, (choice_dna, choice) in enumerate(zip(choices_dnas, choices)):
        if choice_dna.value != choice:
          child_dna = None
        else:
          child_dna = DNA(
              None,
              children=choice_dna.children,
              spec=self.candidates[choice])
        child_dnas.append(child_dna)

    children = []
    for i, c in enumerate(choices):
      children.append(DNA(value=c, children=[
          self.candidates[c].random_dna(
              random_generator, False, child_dnas[i])]))
    return DNA(value=None, children=children)

  def __len__(self) -> int:
    """Returns number of decision points in current space."""
    sub_length = sum([len(c) for c in self.candidates])
    return self.num_choices * (1 + sub_length)

  def __getitem__(
      self, index: Union[int, slice, str]
      ) -> Union[DecisionPoint, List[DecisionPoint]]:
    """Operator [] to return the sub choice(s) if index is int."""
    if isinstance(index, (int, slice)):
      if self.num_choices == 1:
        sub_choices = [self]
      else:
        sub_choices = self._subchoice_specs
      return sub_choices[index]
    return super().__getitem__(index)

  def format(self,
             compact: bool = True,
             verbose: bool = True,
             root_indent: int = 0,
             show_id: bool = True,
             **kwargs):
    """Format this object."""
    if not compact:
      return super().format(compact, verbose, root_indent, **kwargs)

    def _indent(text, indent):
      return ' ' * 2 * indent + text

    s = [f'Choices(num_choices={self.num_choices}, [\n']
    if not self.literal_values:
      for i, candidate in enumerate(self.candidates):
        candidate_str = candidate.format(
            compact, verbose, root_indent + 1, **kwargs)
        s.append(_indent(f'({i}): {candidate_str}\n', root_indent + 1))
    else:
      assert len(self.literal_values) == len(self.candidates)
      for i, (candidate, literal_value) in enumerate(
          zip(self.candidates, self.literal_values)):
        if not candidate.is_constant:
          value_str = candidate.format(compact, verbose, root_indent + 1,
                                       **kwargs)
        else:
          value_str = literal_value
        s.append(_indent(f'({i}): {value_str}\n', root_indent + 1))
    s.append(_indent(']', root_indent))
    if show_id:
      kvlist = [('id', str(self.id), '\'\'')]
    else:
      kvlist = []
    additionl_properties = utils.kvlist_str(
        kvlist
        + [
            ('name', self.name, None),
            ('distinct', self.distinct, True),
            ('sorted', self.sorted, False),
            ('hints', self.hints, None),
            ('subchoice_index', self.subchoice_index, None),
        ],
        compact=False,
        root_indent=root_indent,
    )
    if additionl_properties:
      s.append(', ')
      s.append(additionl_properties)
    s.append(')')
    return ''.join(s)


def manyof(
    num_choices: int,
    candidates: List[DNASpec],
    distinct: bool = True,
    sorted: bool = False,  # pylint: disable=redefined-builtin
    literal_values: Optional[List[Union[str, int, float]]] = None,
    hints: Any = None,
    location: Union[str, utils.KeyPath] = utils.KeyPath(),
    name: Optional[str] = None,
) -> Choices:
  """Returns a multi-choice specification.

  It creates the genotype for :func:`pyglove.manyof`.

  Example::

    spec = pg.geno.manyof(2, [
        pg.geno.constant(),
        pg.geno.constant(),
        pg.geno.oneof([
            pg.geno.constant(),
            pg.geno.constant()
        ])
    ])

  Args:
    num_choices: Number of choices.
    candidates: A list of ``pg.geno.Space`` objects as the candidates.
    distinct: If True, the decisions for the multiple choices should be
      distinct from each other (based on the value of selected indices).
    sorted: If Ture, the decisions returned should be sorted by the
      values of selected indices.
    literal_values: An optional list of string, integer, or float as the
      literal values for the candidates for display purpose.
    hints: An optional hint object.
    location: A ``pg.KeyPath`` object that indicates the location of the
      decision point.
    name: An optional global unique name for identifying this decision
      point.

  Returns:
    A ``pg.geno.Choices`` object.

  See also:

    * :func:`pyglove.geno.constant`
    * :func:`pyglove.geno.space`
    * :func:`pyglove.geno.oneof`
    * :func:`pyglove.geno.floatv`
    * :func:`pyglove.geno.custom`
  """
  normalized_candidates = []
  for c in candidates:
    if isinstance(c, DecisionPoint):
      c = Space([c])
    normalized_candidates.append(c)
  return Choices(num_choices, normalized_candidates, distinct=distinct,
                 sorted=sorted, literal_values=literal_values,
                 hints=hints, location=location, name=name)


def oneof(
    candidates: List[DNASpec],
    literal_values: Optional[List[Union[str, int, float]]] = None,
    hints: Any = None,
    location: Union[str, utils.KeyPath] = utils.KeyPath(),
    name: Optional[str] = None,
) -> Choices:
  """Returns a single choice specification.

  It creates the genotype for :func:`pyglove.oneof`.

  Example::

    spec = pg.geno.oneof([
        pg.geno.constant(),
        pg.geno.oneof([
            pg.geno.constant(),
            pg.geno.constant()
        ])
    ])

  Args:
    candidates: A list of ``pg.geno.Space`` objects as the candidates.
    literal_values: An optional list of string, integer, or float as the
      literal values for the candidates for display purpose.
    hints: An optional hint object.
    location: A ``pg.KeyPath`` object that indicates the location of the
      decision point.
    name: An optional global unique name for identifying this decision
      point.

  Returns:
    A ``pg.geno.Choices`` object.

  See also:

    * :func:`pyglove.geno.constant`
    * :func:`pyglove.geno.space`
    * :func:`pyglove.geno.manyof`
    * :func:`pyglove.geno.floatv`
    * :func:`pyglove.geno.custom`
  """
  return manyof(1, candidates, literal_values=literal_values,
                hints=hints, location=location, name=name)
